#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that helps to register custom validators for the feature types and
extending registered validators with dispatching based on the specific arguments.

Classes
-------
    FeatureValidator
        The Feature Validator class to manage custom validators.
    FeatureValidatorMethod
        The Feature Validator Method class. Extends methods which requires
        dispatching based on the specific arguments.
"""
import inspect
from typing import Any, Callable, Dict, Tuple, Union

import pandas as pd


class WrongHandlerMethodSignature(ValueError):
    def __init__(self, handler_name: str, condition: str, handler_signature: str):
        super().__init__(
            f"The registered condition {condition} is not compatible "
            f"with the provided {handler_name} method. "
            f"Expected parameters: {handler_signature}"
        )


class ValidatorNotFound(ValueError):
    def __init__(self, name: str):
        super().__init__(f"Validator {name} not found.")


class ValidatorWithConditionNotFound(ValueError):
    def __init__(self, name: str):
        super().__init__(f"Validator {name} with provided condition not found.")


class ValidatorAlreadyExists(ValueError):
    def __init__(self, name: str):
        super().__init__(f"Validator {name} already exists.")


class ValidatorWithConditionAlreadyExists(ValueError):
    def __init__(self, name: str):
        super().__init__(f"Validator {name} with provided condition already exists.")


def _prepare_condition(params: Union[Tuple, Dict[str, Any]]) -> Tuple:
    """Converts provided parameters to Tuple.

    Parameters
    -----------
    params: (Union[Tuple, Dict[str, Any]])
        The condition which will be used to register a new validator.

    Returns
    -------
    Tuple
        Prepared condition.

    Raises
    ------
    ValueError
        If condition not provided or provided in the wrong format.
    """
    if not params:
        raise ValueError("Condition not provided.")
    if not isinstance(params, (dict, tuple)):
        raise ValueError(
            "Wrong format for the condition. Condition should be dict or list."
        )
    if not isinstance(params, tuple):
        return tuple((key, params[key]) for key in params)
    return params


class FeatureValidator:
    """The Feature Validator class to manage custom validators.

    Methods
    -------
    register(self, name: str, handler: Callable, condition: Union[Tuple, Dict[str, Any]] = None, replace: bool = False) -> None
        Registers new validator.
    unregister(self, name: str, condition: Union[Tuple, Dict[str, Any]] = None) -> None
        Unregisters validator.
    registered(self) -> pd.DataFrame
        Gets the list of registered validators.


    Examples
    --------
    >>> series = pd.Series(['+1-202-555-0141', '+1-202-555-0142'], name='Phone Number')

    >>> def phone_number_validator(data: pd.Series) -> pd.Series:
    ...    print("phone_number_validator")
    ...    return data

    >>> def universal_phone_number_validator(data: pd.Series, country_code) -> pd.Series:
    ...    print("universal_phone_number_validator")
    ...    return data

    >>> def us_phone_number_validator(data: pd.Series, country_code) -> pd.Series:
    ...    print("us_phone_number_validator")
    ...    return data

    >>> PhoneNumber.validator.register(name="is_phone_number", handler=phone_number_validator, replace=True)
    >>> PhoneNumber.validator.register(name="is_phone_number", handler=universal_phone_number_validator, condition = ('country_code',))
    >>> PhoneNumber.validator.register(name="is_phone_number", handler=us_phone_number_validator, condition = {'country_code':'+1'})

    >>> PhoneNumber.validator.is_phone_number(series)
        phone_number_validator
        0     +1-202-555-0141
        1     +1-202-555-0142

    >>> PhoneNumber.validator.is_phone_number(series, country_code = '+7')
        universal_phone_number_validator
        0     +1-202-555-0141
        1     +1-202-555-0142

    >>> PhoneNumber.validator.is_phone_number(series, country_code = '+1')
        us_phone_number_validator
        0     +1-202-555-0141
        1     +1-202-555-0142

    >>> PhoneNumber.validator.registered()
                   Validator                 Condition                            Handler
        ---------------------------------------------------------------------------------
        0    is_phone_number                        ()             phone_number_validator
        1    is_phone_number          ('country_code')   universal_phone_number_validator
        2    is_phone_number    {'country_code': '+1'}          us_phone_number_validator

    >>> series.ads.validator.is_phone_number()
        phone_number_validator
            0     +1-202-555-0141
            1     +1-202-555-0142

    >>> series.ads.validator.is_phone_number(country_code = '+7')
        universal_phone_number_validator
            0     +1-202-555-0141
            1     +1-202-555-0142

    >>> series.ads.validator.is_phone_number(country_code = '+1')
        us_phone_number_validator
        0     +1-202-555-0141
        1     +1-202-555-0142
    """

    def __init__(self):
        """Initializes the FeatureValidator."""
        self._validators = {}

    def register(
        self,
        name: str,
        handler: Callable,
        condition: Union[Tuple, Dict[str, Any]] = None,
        replace: bool = False,
    ) -> None:
        """Registers new validator.

        Parameters
        ----------
        name : str
            The validator name.
        handler: callable
            The handler.
        condition: Union[Tuple, Dict[str, Any]]
            The condition for the validator.
        replace: bool
            The flag indicating if the registered validator should be replaced with the new one.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            The name is empty or handler is not provided.
        TypeError
            The handler is not callable.
            The name of the validator is not a string.
        ValidatorAlreadyExists
            The validator is already registered.
        """
        if not name:
            raise ValueError("Validator name is not provided.")
        if not isinstance(name, str):
            raise TypeError("Validator name should be a string.")
        if not replace and name in self._validators:
            if not condition:
                raise ValidatorAlreadyExists(name)
            if self._validators[name]._has_condition(condition):
                raise ValidatorWithConditionAlreadyExists(name)
        if not handler:
            raise ValueError("Handler is not provided.")
        if not callable(handler):
            raise TypeError("Handler should be a function.")

        if condition:
            self._validators[name].register(condition=condition, handler=handler)
        else:
            self._validators[name] = FeatureValidatorMethod(handler)

    def unregister(
        self, name: str, condition: Union[Tuple, Dict[str, Any]] = None
    ) -> None:
        """Unregisters validator.

        Parameters
        ----------
        name: str
            The name of the validator to be unregistered.
        condition: Union[Tuple, Dict[str, Any]]
            The condition for the validator to be unregistered.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        TypeError
            The name of the validator is not a string.
        ValidatorNotFound
            The validator not found.
        ValidatorWIthConditionNotFound
            The validator with provided condition not found.
        """
        if not name:
            raise ValueError("Validator name is not provided.")
        if not isinstance(name, str):
            raise TypeError("Validator name should be a string.")
        if name not in self._validators:
            raise ValidatorNotFound(name)
        if condition and not self._validators[name]._has_condition(condition):
            raise ValidatorWithConditionNotFound(name)

        if condition:
            self._validators[name].unregister(condition)
        else:
            del self._validators[name]

    def registered(self) -> pd.DataFrame:
        """Gets the list of registered validators.

        Returns
        -------
        pd.DataFrame
            The list of registerd validators.
        """
        result_df = pd.DataFrame((), columns=["Validator", "Condition", "Handler"])
        for key, feature_validator in self._validators.items():
            feature_validators_df = feature_validator.registered()
            feature_validators_df.insert(0, "Validator", key)
            result_df = result_df.append(feature_validators_df)
        result_df.reset_index(drop=True, inplace=True)
        return result_df

    def _bind_data(self, data: pd.Series) -> None:
        """Binds the data to the all registered validators.

        Parameters
        ----------
        data: pd.Series
            The data to be processed.
        """
        for validator in self._validators.values():
            validator._bind_data(data)

    def __getattr__(self, attr):
        """Makes it possible to invoke registered validators as a regular method."""
        if attr in self._validators:
            return self._validators[attr]
        raise AttributeError(attr)


class FeatureValidatorMethod:
    """The Feature Validator Method class.

    Extends methods which requires dispatching based on the specific arguments.

    Methods
    -------
    register(self, condition: Union[Tuple, Dict[str, Any]], handler: Callable) -> None
        Registers new handler.
    unregister(self, condition: Union[Tuple, Dict[str, Any]]) -> None
        Unregisters existing handler.
    registered(self) -> pd.DataFrame
        Gets the list of registered handlers.
    """

    def __init__(self, handler: Callable):
        """Initializes the Feature Validator Method.

        Parameters
        ----------
        handler: Callable
            The handler that will be called by default if suitable one not found.
        """
        if not handler:
            raise ValueError("Default handler is not specified.")

        self._default_handler = handler
        self._handlers = {}
        self._data = None

    def register(
        self, condition: Union[Tuple, Dict[str, Any]], handler: Callable
    ) -> None:
        """Registers new handler.

        Parameters
        -----------
        condition: Union[Tuple, Dict[str, Any]]
            The condition which will be used to register a new handler.
        handler: Callable
            The handler to be registered.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If condition not provided or provided in the wrong format.
            If handler not provided or has wrong format.
        """
        if not condition:
            raise ValueError("Condition not provided.")
        if not isinstance(condition, (dict, tuple)):
            raise ValueError(
                "Wrong format for the condition. Condition should be dict or list."
            )
        if not handler or not callable(handler):
            raise ValueError("Handler not provided. Handler should be a function.")

        prepared_condition = _prepare_condition(condition)
        # self.__validate_handler_signature(handler)
        self._handlers[prepared_condition] = handler
        self._data = None

    def unregister(self, condition: Union[Tuple, Dict[str, Any]]) -> None:
        """Unregisters existing handler.

        Parameters
        -----------
        condition: Union[Tuple, Dict[str, Any]]
            The condition which will be used to unregister a handler.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If condition not provided or provided in the wrong format.
            If condition not registered.
        """
        if not condition:
            raise ValueError("Condition not provided.")
        if not isinstance(condition, (dict, tuple)):
            raise ValueError(
                "Wrong format for the condition. Condition should be dict or list."
            )
        prepared_condition = _prepare_condition(condition)
        if prepared_condition not in self._handlers:
            raise ValueError("Condition not registered.")
        del self._handlers[prepared_condition]

    def registered(self) -> pd.DataFrame:
        """Gets the list of registered handlers.

        Returns
        -------
        pd.DataFrame
            The list of registerd handlers.
        """
        result = [("()", self._default_handler.__name__)]
        for key, value in self._handlers.items():
            try:
                str_key = str(dict(key))
            except ValueError:
                str_key = str(key)
            result.append((str_key, value.__name__))
        return pd.DataFrame(result, columns=["Condition", "Handler"])

    def _process(self, *args, **kwargs) -> pd.Series:
        """Finds and invokes a suitable handler for the provided condition.

        Parameters
        ----------
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments. Parameters to search suitable handler.

        Returns
        -------
        pd.Series
            The result of invoked handler.
        """
        if kwargs:
            for key in (
                tuple((key, kwargs[key]) for key in kwargs),
                tuple(kwargs.keys()),
            ):
                if key in self._handlers:
                    return self._handlers[key](self._data, *args, **kwargs)

        return self._default_handler(self._data, *args, **kwargs)

    def _bind_data(self, data: pd.Series) -> None:
        """Binds the data to the validator.

        Parameters
        ----------
        data: pd.Series
            The data to be processed.
        """
        self._data = data

    def _validate_handler_signature(
        self, condition: Union[Tuple, Dict[str, Any]], handler: Callable
    ) -> bool:
        """Validates handler signature.

        Parameters
        ----------
        condition: Union[Tuple, Dict[str, Any]]
            The condition to validate.
        handler: Callabe
            The hanlder to validate.

        Returns
        -------
        bool
            True if provided condition and handler arguments compatible.

        Raises
        -------
        WrongHandlerMethodSignature
            If provided condition and handler arguments not compatible.
        """
        prepared_condition = _prepare_condition(condition)
        handler_args = list(inspect.signature(handler).parameters.keys())
        params_args = ["data"] + (
            list(arg[0] for arg in prepared_condition)
            if isinstance(prepared_condition[0], tuple)
            else list(prepared_condition)
        )
        if handler_args != params_args:
            raise WrongHandlerMethodSignature(
                handler.__name__, str(params_args), str(handler_args)
            )
        return True

    def _has_condition(self, condition: Union[Tuple, Dict[str, Any]]) -> bool:
        """Checks whether provided condition registered or not.

        Parameters
        ----------
        condition: Union[Tuple, Dict[str, Any]]
                The condition to check.
        Returns
        -------
        bool
            True if condition registered, False othervise.

        Raises
        ------
        ValueError
            If condition not provided or has wrong format.
        """
        if not condition:
            raise ValueError("Condition not provided.")
        if not isinstance(condition, (dict, tuple)):
            raise ValueError(
                "Wrong format for the condition. Condition should be dict or list."
            )
        prepared_condition = _prepare_condition(condition)
        return prepared_condition in self._handlers

    def __call__(self, *args, **kwargs) -> pd.Series:
        """Makes class instance callable.

        Parameters
        ----------
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.

        Returns
        -------
        pd.Series
            The result of processing data.

        Raises
        ------
        ValueError
            If data is not provided,
        TypeError
            If data has wrong format.
        """
        if args and len(args) > 0:
            self._data = args[0]
        if self._data is None:
            raise ValueError("Data is not provided.")
        if not isinstance(self._data, pd.Series):
            raise TypeError("Wrong data format. Data should be Series.")
        return self._process(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from tempfile import gettempprefix
import fsspec
import yaml
import ads
import os
import oci
import json
from ads.jobs import Job
import ipaddress
from abc import ABCMeta, abstractmethod
import subprocess
from time import sleep, time
import pandas as pd  # Have to find a better way for timedelta
from urllib.parse import urlparse
from ads.opctl.utils import suppress_traceback


class AbstractOperatorSpec(metaclass=ABCMeta):
    YAML_TRANSLATION = {
        "docker_image": NotImplementedError,
    }

    def __init__(self, yaml_file, **kwargs):
        # TODO: Move into factory method, in order to determine OpSpec instance
        self.debug = kwargs["debug"]
        self.spec = _ingest_yaml_file(yaml_file, self.debug)
        if yaml_file:
            if os.path.exists(yaml_file):
                with open(yaml_file, "r") as f:
                    self.spec = suppress_traceback(self.debug)(yaml.safe_load)(f.read())
            else:
                raise FileNotFoundError(f"{yaml_file} is not found")
        else:
            self.spec = {}

        # TODO: init starts here:

        # Validate user yaml input
        # TODO: ask infra group: merge in additional args from cli, env vars, defaults?
        # Or should yaml files be forced to be completed, and not overwritten (
        # maybe there's lee-way? backend can change from cli or something?)
        # then re-validate?

        # Need todo equivalent of:
        # p = ConfigProcessor(config).step(ConfigMerger, **kwargs)

        # p.step(ConfigResolver).step(ConfigValidator)
        # # spec may have changed during validation step (e.g. defaults filled in)
        # # thus command need to be updated since it encodes spec
        # p = ConfigResolver(p.config)
        # p._resolve_command()
        # self.spec = p.config

    @staticmethod
    def build(yaml_file, **kwargs):
        return AbstractOperatorSpec(yaml_file, **kwargs)

    def __getitem__(self, key):
        # Note, we have not written a __setitem__ method because we do not want users setting these outside of the class
        return self._translate_key(key)

    def _translate_key(self, key):
        key_location = self.YAML_TRANSLATION[key]
        return self.spec.__getitem__(key_location)

    def get_keys(self):
        return self.YAML_TRANSLATION.keys()

    @property
    def docker_image(self):
        pass


class V0Spec(AbstractOperatorSpec):
    YAML_TRANSLATION = {
        "docker_image": ["execution"]["image"],
        "backend": ["execution"]["backend"],
        "operator_slug": ["execution"]["operator_slug"],
        "debug": ["execution"]["debug"],
        "operator_folder_path": ["execution"]["operator_folder_path"],
        "auth": ["execution"]["auth"],
        "oci_profile": ["execution"]["oci_profile"],
        "oci_config": ["execution"]["oci_config"],
        "conda_pack_folder": ["execution"]["conda_pack_folder"],
        "conda_pack_os_prefix": ["execution"]["conda_pack_os_prefix"],
        "compartment_id": ["infrastructure"]["compartment_id"],
        "project_id": ["infrastructure"]["project_id"],
        "subnet_id": ["infrastructure"]["subnet_id"],
        "log_group_id": ["infrastructure"]["log_group_id"],
        "log_id": ["infrastructure"]["log_id"],
        "shape_name": ["infrastructure"]["shape_name"],
        "block_storage_size_in_GBs": ["infrastructure"]["block_storage_size_in_GBs"],
        "docker_registry": ["infrastructure"]["docker_registry"],
    }

    @staticmethod
    def build(yaml_file, **kwargs):
        return V0Spec(yaml_file, **kwargs)


class DistributedV1Spec(AbstractOperatorSpec):
    # TODO: move this to a JobsClusterInfra spec class
    YAML_TRANSLATION = {
        # framework
        "framework_config_name": ["spec"]["Framework"]["spec"]["frameworkConfigName"],
        "framework_config": ["spec"]["Framework"]["spec"]["frameworkConfig"],
        "nanny_port_range": ["spec"]["Framework"]["spec"]["schedulerConfig"][
            "nannyPortRange"
        ],
        "worker_port_range": ["spec"]["Framework"]["spec"]["schedulerConfig"][
            "workerPortRange"
        ],
        "timeout": ["spec"]["Framework"]["spec"]["workerConfig"]["timeout"],
        "work_dir": ["spec"]["Framework"]["spec"]["workDir"],
        # Runtime
        "entrypoint": ["spec"]["Runtime"]["spec"]["entrypoint"],
        "args": ["spec"]["Runtime"]["spec"]["args"],
        "environmentVariables": ["spec"]["Runtime"]["spec"]["environmentVariables"],
        # Infra
        "docker_image": ["spec"]["Infrastructure"]["spec"]["dockerImage"],
        "display_name": ["spec"]["Infrastructure"]["spec"]["displayName"],
        "compartment_id": ["spec"]["Infrastructure"]["spec"]["compartmentId"],
        "project_id": ["spec"]["Infrastructure"]["spec"]["projectId"],
        "subnet_id": ["spec"]["Infrastructure"]["spec"]["subnetId"],
        "log_group_id": ["spec"]["Infrastructure"]["spec"]["logGroupId"],
        "log_id": ["spec"]["Infrastructure"]["spec"]["logId"],
        "shape_name": ["spec"]["Infrastructure"]["spec"]["cluster"]["shapeName"],
        "block_storage_size_in_GBs": ["spec"]["Infrastructure"]["spec"]["cluster"][
            "blockStorageSizeGB"
        ],
        "num_workers": ["spec"]["Infrastructure"]["spec"]["cluster"]["numWorkers"],
        "ephemeral": ["spec"]["Infrastructure"]["spec"]["cluster"]["ephemeral"],
    }

    @staticmethod
    def build(yaml_file, **kwargs):
        return DistributedV1Spec(yaml_file, **kwargs)


def _ingest_yaml_file(yaml_file, debug=False):
    if yaml_file:
        if os.path.exists(yaml_file):
            with open(yaml_file, "r") as f:
                spec = suppress_traceback(debug)(yaml.safe_load)(f.read())
        else:
            raise FileNotFoundError(f"{yaml_file} is not found")
    else:
        spec = {}
    return spec



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a Continuous feature type.

Classes:
    Continuous
        The Continuous feature type.
"""
import matplotlib.pyplot as plt
import pandas as pd
from ads.feature_engineering.feature_type.base import FeatureType
from ads.feature_engineering.utils import (
    _add_missing,
    _set_seaborn_theme,
    SchemeTeal,
    _format_stat,
)
from ads.feature_engineering import schema
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class Continuous(FeatureType):
    """
    Type representing continuous values.

    Attributes
    ----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.
    feature_plot(x: pd.Series) -> plt.Axes
        Shows distributions of datasets using box plot.
    """

    description = "Type representing continuous values."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, mean, standard deviation, sample minimum,
        lower quartile, median, 75%, upper quartile, skew and missing(count).

        Examples
        --------
        >>> cts = pd.Series([13.32, 3.32, 4.3, 2.45, 6.34, 2.25,
                            4.43, 3.26, np.NaN, None], name='continuous')
        >>> cts.ads.feature_type = ['continuous']
        >>> cts.ads.feature_stat()
            Metric                  Value
        0	count	                10.000
        1	mean	                4.959
        2	standard deviation	    3.620
        3	sample minimum	        2.250
        4	lower quartile	        3.058
        5	median	                3.810
        6	upper quartile	        4.908
        7	sample maximum	        13.320
        8	skew	                2.175
        9	missing	                2.000

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.
        """
        df_stat = x.describe()
        _format_stat(df_stat)
        df_stat["count"] = len(x)
        df_stat["skew"] = x.skew()
        df_stat = _add_missing(x, df_stat).to_frame()
        df_stat.iloc[:, 0] = df_stat.iloc[:, 0].round(3)
        return df_stat

    @staticmethod
    @runtime_dependency(module="seaborn", install_from=OptionalDependency.VIZ)
    def feature_plot(x: pd.Series) -> plt.Axes:
        """
        Shows distributions of datasets using box plot.

        Examples
        --------
        >>> cts = pd.Series([13.32, 3.32, 4.3, 2.45, 6.34, 2.25,
                            4.43, 3.26, np.NaN, None], name='continuous')
        >>> cts.ads.feature_type = ['continuous']
        >>> cts.ads.feture_plot()

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
            Plot object for the series based on the Continuous feature type.
        """
        col_name = x.name if x.name else "continuous"
        df = x.to_frame(name=col_name)
        df = df[pd.to_numeric(df[col_name], errors="coerce").notnull()]
        if len(df.index):
            _set_seaborn_theme()
            return seaborn.boxplot(
                x=df[col_name], width=0.2, color=SchemeTeal.AREA_DARK
            )

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> cts = pd.Series([13.32, 3.32, 4.3, 2.45, 6.34, 2.25,
                            4.43, 3.26, np.NaN, None], name='continuous')
        >>> cts.ads.feature_type = ['continuous']
        >>> cts.ads.feature_domain()
        constraints: []
        stats:
            count: 10.0
            lower quartile: 3.058
            mean: 4.959
            median: 3.81
            missing: 2.0
            sample maximum: 13.32
            sample minimum: 2.25
            skew: 2.175
            standard deviation: 3.62
            upper quartile: 4.908
        values: Continuous

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the Continuous feature type.
        """

        return schema.Domain(cls.__name__, cls.feature_stat(x).to_dict()[x.name], [])



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import numpy as np
import pandas as pd

from ads.common import logger
from ads.dataset.exception import ValidationError
from collections import defaultdict, Counter
from itertools import product, combinations
from typing import Tuple


def _cat_vs_cat(df: pd.core.frame.DataFrame, categorical_columns: list) -> pd.DataFrame:
    """
    calc the correlation of all pairs of categorical features and categorical features
    """
    if not categorical_columns:
        return pd.DataFrame()
    categorical_pairs = list(combinations(categorical_columns, 2))
    corr_list = []
    for col in categorical_pairs:
        cat1_name = col[0]
        cat2_name = col[1]
        _check_if_same_type(df[col[0]], cat1_name)
        _check_if_same_type(df[col[1]], cat2_name)
        corr_list.append(
            _cramers_v(np.array(df[col[0]].values), np.array(df[col[1]].values))
        )
    correlation_matrix = _list_to_dataframe(categorical_pairs, corr_list)
    return correlation_matrix


def _cat_vs_cts(
    df: pd.core.frame.DataFrame, categorical_columns: list, continuous_columns: list
) -> pd.DataFrame:
    """
    calc the correlation of all pairs of categorical features and continuous features
    """
    numerical_categorical_pairs = list(product(categorical_columns, continuous_columns))
    corr_list = []
    for col in numerical_categorical_pairs:
        corr_list.append(
            _correlation_ratio(np.array(df[col[0]].values), np.array(df[col[1]].values))
        )
    correlation_matrix = _list_to_dataframe(numerical_categorical_pairs, corr_list)
    return correlation_matrix


def _list_to_dataframe(name_list: list, corr_list: list) -> pd.DataFrame:
    corr_dict = defaultdict(dict)
    for idx, corr in zip(name_list, corr_list):
        row_name = idx[0]
        col_name = idx[1]
        corr_dict[row_name][col_name] = corr_dict[col_name][row_name] = round(corr, 4)
        corr_dict[row_name][row_name] = corr_dict[col_name][col_name] = 1.0000
    correlation_matrix = pd.DataFrame.from_dict(corr_dict).sort_index()
    correlation_matrix = correlation_matrix.loc[:, correlation_matrix.index]
    return correlation_matrix


def _correlation_ratio(cat: np.ndarray, cts: np.ndarray):
    """
    calc the correlation of a pair of a categorical feature and a continuous feature
    using correlation ratio when input are two numpy arrays
    """
    keep_cts = ~pd.isnull(cts)
    cat_no_nan = cat[keep_cts]
    cts_no_nan = cts[keep_cts]

    keep_cat = ~pd.isnull(cat_no_nan)
    cat_no_none = cat_no_nan[keep_cat]
    cts_no_none = cts_no_nan[keep_cat]

    unq_cat, tags, group_count = np.unique(
        list(cat_no_none), return_inverse=1, return_counts=1
    )
    group_mean = np.bincount(tags, cts_no_none) / group_count
    overall_mean = np.nanmean(cts_no_none)
    n = len(cts_no_none)

    dispersion_within = np.dot(group_count, np.square(group_mean - overall_mean))
    dispersion_population = cts_no_none.var() * n
    ratio = dispersion_within / dispersion_population

    return np.sqrt(ratio)


def _count_occurrence(
    cat1: np.ndarray, cat2: np.ndarray
) -> Tuple[np.ndarray, int, int]:
    """
    calc the contingency table of two arrays
    """
    occurance_cnt = Counter([(x, y) for x, y in zip(cat1, cat2)])
    nunique_cat1 = np.unique(cat1[~pd.isnull(cat1)])
    nunique_cat2 = np.unique(cat2[~pd.isnull(cat2)])
    r = len(nunique_cat1)
    k = len(nunique_cat2)
    contigency_table = np.zeros((r, k))
    for row, num1 in enumerate(nunique_cat1):
        for col, num2 in enumerate(nunique_cat2):
            contigency_table[row, col] = occurance_cnt[(num1, num2)]

    return contigency_table, r, k


def _chi_squared(count_matrix: np.ndarray, n_obs: int) -> float:
    """
    Compute Chi-squared when given a contingency table
    """
    row_sums = np.tile(np.sum(count_matrix, axis=1), (count_matrix.shape[1], 1)).T
    col_sums = np.tile(np.sum(count_matrix, axis=0), (count_matrix.shape[0], 1))
    return np.sum(
        np.square(count_matrix - row_sums * col_sums / n_obs)
        / (row_sums * col_sums / n_obs)
    )


def _cramers_v(cat1: np.ndarray, cat2: np.ndarray) -> float:
    """
    calc the cramers v of two numpy arrays
    """
    n = len(cat1)
    if n == 1:
        return 0
    contigency_table, r, k = _count_occurrence(cat1, cat2)

    if r == 0:
        return 0.0000

    chi2 = _chi_squared(contigency_table, n)
    phi2 = chi2 / n

    phi2corr = max(0, phi2 - ((k - 1) * (r - 1)) / (n - 1))
    rcorr = r - (np.square(r - 1)) / (n - 1)
    kcorr = k - (np.square(k - 1)) / (n - 1)
    denominator = min((kcorr - 1), (rcorr - 1))
    if denominator == 0:
        return np.nan
    return np.sqrt(phi2corr / denominator)


def _get_columns_by_type(
    feature_types_df: pd.DataFrame, threshold: float = 0.8
) -> Tuple[list, list, list]:
    """
    return the categorical columns, continuous columns and columns of other types
    """
    missing = feature_types_df.loc[:, "missing_percentage"] > threshold
    not_missing = feature_types_df.loc[:, "missing_percentage"] <= threshold
    missing_columns = list(feature_types_df.loc[missing, "feature_name"].values)
    constant_columns = list(
        feature_types_df.loc[
            (feature_types_df.loc[:, "type"].isin(["constant"])), "feature_name"
        ].values
    )
    categorical_columns = list(
        feature_types_df.loc[
            (feature_types_df.loc[:, "type"].isin(["categorical", "zipcode"]))
            & not_missing,
            "feature_name",
        ].values
    )
    continuous_columns = list(
        feature_types_df.loc[
            (feature_types_df.loc[:, "type"].isin(["continuous", "ordinal"]))
            & not_missing,
            "feature_name",
        ].values
    )
    other_columns = list(
        set(feature_types_df.index.values)
        - set(categorical_columns)
        - set(continuous_columns)
        - set(missing_columns)
        - set(constant_columns)
    )

    if missing_columns:
        logger.info(
            f"The columns {missing_columns} are not included because more than {threshold}% of the values are missing. "
        )
        logger.info(
            f"The columns {missing_columns} are not included because more than {threshold}% of the values are missing. "
            f"Adjust this threshold using the `nan_threshold` parameter."
        )
    if constant_columns:
        logger.info(
            " The constant columns {} are not included.".format(constant_columns)
        )
    if other_columns:
        logger.info(
            f" The columns {other_columns} are not included because more than {threshold}% of the values are missing, or "
            f"they are not one of the following types: "
            f"`categorical`, `zipcode`, `continuous`"
            f", or `ordinal`."
        )

    return categorical_columns, continuous_columns, other_columns


def _validate_correlation_methods(correlation_methods):
    if isinstance(correlation_methods, str):
        correlation_methods = [correlation_methods]
    for method in correlation_methods:
        if method not in ["all", "pearson", "cramers v", "correlation ratio"]:
            raise ValidationError(f"{method} is not supported.")
    if "all" in [method for method in correlation_methods]:
        correlation_methods = ["pearson", "cramers v", "correlation ratio"]
    return correlation_methods


def _check_if_same_type(series, col_name):
    col = series.dropna().values
    if len(col) > 0:
        col_type = type(col[0])
        if not all([isinstance(x, col_type) for x in col]):
            raise TypeError(
                f"More than one data type in the column `{col_name}`. Keep all the values in that column the same type."
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from collections import defaultdict
from typing import Callable, DefaultDict, List, Optional, cast

import matplotlib.pyplot as plt
import numpy as np
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)

try:
    from optuna.visualization.matplotlib._matplotlib_imports import _imports
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        f"The `optuna` module was not found. Please run `pip install "
        f"{OptionalDependency.OPTUNA}`."
    )
except Exception as e:
    raise

if _imports.is_successful():
    from optuna.visualization.matplotlib._matplotlib_imports import (
        Axes,
        LineCollection,
        plt,
    )


@runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
def plot_parallel_coordinate(
    study: "optuna.study.Study",
    params: Optional[List[str]] = None,
    *,
    target: Optional[Callable[["optuna.trial.FrozenTrial"], float]] = None,
    target_name: str = "Objective Value",
) -> "Axes":
    """Plot the high-dimensional parameter relationships in a study with Matplotlib."""

    _imports.check()
    optuna.visualization._utils._check_plot_args(study, target, target_name)
    return _get_parallel_coordinate_plot(study, params, target, target_name)


@runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
def _get_parallel_coordinate_plot(
    study: "optuna.study.Study",
    params: Optional[List[str]] = None,
    target: Optional[Callable[["optuna.trial.FrozenTrial"], float]] = None,
    target_name: str = "Objective Value",
    fig_size=None,
) -> "Axes":

    if target is None:

        def _target(t: "optuna.trial.FrozenTrial") -> float:
            return cast(float, t.value)

        target = _target
        reversescale = (
            study.direction == optuna.study._study_direction.StudyDirection.MINIMIZE
        )
    else:
        reversescale = True

    # Set up the graph style.
    fig, ax = plt.subplots()
    cmap = plt.get_cmap("Blues_r" if reversescale else "Blues")
    ax.set_title("Parallel Coordinate Plot")
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)

    # Prepare data for plotting.
    trials = [
        trial
        for trial in study.trials
        if trial.state == optuna.trial.TrialState.COMPLETE
    ]

    if len(trials) == 0:
        logger.warning("Your study does not have any completed trials.")
        return ax

    all_params = {p_name for t in trials for p_name in t.params.keys()}
    if params is not None:
        for input_p_name in params:
            if input_p_name not in all_params:
                raise ValueError(
                    "Parameter {} does not exist in your study.".format(input_p_name)
                )
        all_params = set(params)
    sorted_params = sorted(all_params)

    obj_org = [target(t) for t in trials]
    obj_min = min(obj_org)
    obj_max = max(obj_org)
    obj_w = obj_max - obj_min
    dims_obj_base = [[o] for o in obj_org]

    cat_param_names = []
    cat_param_values = []
    cat_param_ticks = []
    param_values = []
    var_names = [target_name]
    for p_name in sorted_params:
        values = [t.params[p_name] if p_name in t.params else np.nan for t in trials]

        from optuna.visualization.matplotlib._utils import _is_categorical

        if _is_categorical(trials, p_name):
            vocab = defaultdict(lambda: len(vocab))  # type: DefaultDict[str, int]
            values = [vocab[v] for v in values]
            cat_param_names.append(p_name)
            vocab_item_sorted = sorted(vocab.items(), key=lambda x: x[1])
            cat_param_values.append([v[0] for v in vocab_item_sorted])
            cat_param_ticks.append([v[1] for v in vocab_item_sorted])

        p_min = min(values)
        p_max = max(values)
        p_w = p_max - p_min

        if p_w == 0.0:
            center = obj_w / 2 + obj_min
            for i in range(len(values)):
                dims_obj_base[i].append(center)
        else:
            for i, v in enumerate(values):
                dims_obj_base[i].append((v - p_min) / p_w * obj_w + obj_min)

        var_names.append(p_name if len(p_name) < 20 else "{}...".format(p_name[:17]))
        param_values.append(values)

    # Draw multiple line plots and axes.
    # Ref: https://stackoverflow.com/a/50029441
    ax.set_xlim(0, len(sorted_params))
    ax.set_ylim(obj_min, obj_max)
    xs = [range(len(sorted_params) + 1) for _ in range(len(dims_obj_base))]
    segments = [np.column_stack([x, y]) for x, y in zip(xs, dims_obj_base)]
    lc = LineCollection(segments, cmap=cmap)
    lc.set_array(np.asarray([target(t) for t in trials] + [0]))
    axcb = fig.colorbar(lc, pad=0.1, ax=fig.gca())

    axcb.set_label(target_name)
    plt.xticks(range(len(sorted_params) + 1), var_names, rotation=330)

    for i, p_name in enumerate(sorted_params):
        ax2 = ax.twinx()
        ax2.set_ylim(min(param_values[i]), max(param_values[i]))

        ax2.spines["top"].set_visible(False)
        ax2.spines["bottom"].set_visible(False)
        ax2.get_xaxis().set_visible(False)
        ax2.plot([1] * len(param_values[i]), param_values[i], visible=False)
        ax2.spines["right"].set_position(("axes", (i + 1) / len(sorted_params)))
        if p_name in cat_param_names:
            idx = cat_param_names.index(p_name)
            tick_pos = cat_param_ticks[idx]
            tick_labels = cat_param_values[idx]
            ax2.set_yticks(tick_pos)
            ax2.set_yticklabels(tick_labels)

    ax.add_collection(lc)
    plt.show(block=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging


logger = logging.getLogger(__name__)

from .mlx_global_explainer import *
from .mlx_local_explainer import *
from .mlx_whatif_explainer import *

try:
    import mlx

    mlx.initjs()
except:
    pass



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
import configparser
import click

import ads

from typing import Dict, List

from ads.common.auth import AuthContext
from ads.common.oci_datascience import DSCNotebookSession
from ads.common.extended_enum import ExtendedEnumMeta
from ads.opctl.backend.ads_ml_job import MLJobBackend, MLJobDistributedBackend
from ads.opctl.backend.ads_model_deployment import ModelDeploymentBackend
from ads.opctl.backend.local import (
    LocalBackend,
    LocalBackendDistributed,
    LocalPipelineBackend,
)
from ads.opctl.backend.ads_dataflow import DataFlowBackend
from ads.opctl.backend.ads_ml_pipeline import PipelineBackend
from ads.opctl.config.base import ConfigProcessor
from ads.opctl.config.merger import ConfigMerger
from ads.opctl.config.resolver import ConfigResolver
from ads.opctl.config.utils import read_from_ini
from ads.opctl.config.validator import ConfigValidator
from ads.opctl.config.yaml_parsers import YamlSpecParser
import fsspec
from ads.common.auth import AuthType

from ads.opctl.distributed.cmds import (
    update_ini,
    increment_tag_in_ini,
    docker_build_cmd,
    update_image,
    verify_and_publish_image,
    update_config_image,
)
from ads.opctl.constants import (
    DEFAULT_OCI_CONFIG_FILE,
    DEFAULT_PROFILE,
    DEFAULT_CONDA_PACK_FOLDER,
    DEFAULT_ADS_CONFIG_FOLDER,
    ADS_CONFIG_FILE_NAME,
    ADS_JOBS_CONFIG_FILE_NAME,
    ADS_DATAFLOW_CONFIG_FILE_NAME,
    ADS_ML_PIPELINE_CONFIG_FILE_NAME,
    ADS_LOCAL_BACKEND_CONFIG_FILE_NAME,
    BACKEND_NAME,
)
from ads.opctl.utils import (
    is_in_notebook_session,
    get_service_pack_prefix,
)
import yaml


class DataScienceResource(str, metaclass=ExtendedEnumMeta):
    JOB = "datasciencejob"
    DATAFLOW = "dataflowapplication"
    PIPELINE = "datasciencepipeline"
    MODEL_DEPLOYMENT = "datasciencemodeldeployment"


class DataScienceResourceRun(str, metaclass=ExtendedEnumMeta):
    JOB_RUN = "datasciencejobrun"
    DATAFLOW_RUN = "dataflowrun"
    PIPELINE_RUN = "datasciencepipelinerun"
    MODEL_DEPLOYMENT = "datasciencemodeldeployment"


DATA_SCIENCE_RESOURCE_BACKEND_MAP = {
    DataScienceResource.JOB: "job",
    DataScienceResourceRun.JOB_RUN: "job",
    DataScienceResource.DATAFLOW: "dataflow",
    DataScienceResourceRun.DATAFLOW_RUN: "dataflow",
    DataScienceResource.PIPELINE: "pipeline",
    DataScienceResourceRun.PIPELINE_RUN: "pipeline",
    DataScienceResourceRun.MODEL_DEPLOYMENT: "deployment",
}

DATA_SCIENCE_RESOURCE_RUN_BACKEND_MAP = {
    DataScienceResourceRun.JOB_RUN: "job",
    DataScienceResourceRun.DATAFLOW_RUN: "dataflow",
    DataScienceResourceRun.PIPELINE_RUN: "pipeline",
    DataScienceResourceRun.MODEL_DEPLOYMENT: "deployment",
}


class _BackendFactory:
    BACKENDS_MAP = {
        BACKEND_NAME.JOB.value: MLJobBackend,
        BACKEND_NAME.DATAFLOW.value: DataFlowBackend,
        BACKEND_NAME.PIPELINE.value: PipelineBackend,
        BACKEND_NAME.MODEL_DEPLOYMENT.value: ModelDeploymentBackend,
    }

    LOCAL_BACKENDS_MAP = {
        BACKEND_NAME.JOB.value: LocalBackend,
        BACKEND_NAME.PIPELINE.value: LocalPipelineBackend,
    }

    def __init__(self, config: Dict):
        self.config = config
        self._backend = config["execution"].pop("backend", None)
        if self._backend is None:
            raise RuntimeError("Please specify backend.")
        elif (
            self._backend != BACKEND_NAME.LOCAL.value
            and self._backend not in self.BACKENDS_MAP
        ):
            raise NotImplementedError(f"backend {self._backend} is not implemented.")

    @property
    def backend(self):
        if self._backend == BACKEND_NAME.LOCAL.value:
            kind = self.config.get("kind")
            if kind not in self.LOCAL_BACKENDS_MAP:
                options = [backend for backend in self.LOCAL_BACKENDS_MAP.keys()]
                # Special case local backend option not supported by this factory.
                options.append("distributed")
                raise RuntimeError(
                    f"kind {kind} not supported by local backend. Please choose from: "
                    f"[{str.join('|', options)}]"
                )
            return self.LOCAL_BACKENDS_MAP[kind](self.config)

        return self.BACKENDS_MAP[self._backend](self.config)


def _save_yaml(yaml_content, **kwargs):
    """Saves job run info YAML to a local file.

    Parameters
    ----------
    yaml_content : str
        YAML content as string.
    """
    if kwargs["job_info"]:
        yaml_path = os.path.abspath(os.path.expanduser(kwargs["job_info"]))
        if os.path.isfile(yaml_path):
            overwrite = input(
                f"File {yaml_path} already exists. Overwrite the file? [yN]: "
            )
            if overwrite not in ["y", "Y"]:
                return
        with open(yaml_path, "w", encoding="utf-8") as f:
            f.write(yaml_content)
        print(f"Job run info saved to {yaml_path}")


def run(config: Dict, **kwargs) -> Dict:
    """
    Run a job given configuration and command line args passed in (kwargs).

    Parameters
    ----------
    config: dict
        dictionary of configurations
    kwargs: dict
        keyword arguments, stores configuration from command line args

    Returns
    -------
    Dict
        dictionary of job id and run id in case of ML Job run, else empty if running locally
    """
    p = ConfigProcessor(config).step(ConfigMerger, **kwargs)
    if config.get("kind") == "distributed":  # TODO: add kind factory
        print(
            "......................... Initializing the process ..................................."
        )
        ini = update_ini(
            kwargs["tag"],
            kwargs["registry"],
            kwargs["dockerfile"],
            kwargs["source_folder"],
            config,
            kwargs["nobuild"],
        )
        nobuild = kwargs["nobuild"]
        mode = kwargs["backend"]
        increment = kwargs["auto_increment"]

        if not nobuild:
            if increment:
                ini = increment_tag_in_ini(ini)
            docker_build_cmd(ini)
        config = update_image(config, ini)

        if mode == BACKEND_NAME.LOCAL.value:
            print(
                "\u26A0 Docker Image: "
                + ini.get("main", "registry")
                + ":"
                + ini.get("main", "tag")
                + " is not pushed to oci artifacts."
            )
            print("running image: " + config["spec"]["cluster"]["spec"]["image"])

            backend = LocalBackendDistributed(config)
            backend.run()
        elif mode == BACKEND_NAME.DATAFLOW.value:
            raise RuntimeError(
                "backend operator for distributed training can either be local or job"
            )
        else:
            if not kwargs["dry_run"]:
                verify_and_publish_image(kwargs["nopush"], config)
                print("running image: " + config["spec"]["cluster"]["spec"]["image"])
            cluster_def = YamlSpecParser.parse_content(config)

            backend = MLJobDistributedBackend(p.config)

            # Define job first,
            # Then Run
            cluster_run_info = backend.run(
                cluster_info=cluster_def, dry_run=p.config["execution"].get("dry_run")
            )
            if cluster_run_info:
                cluster_run = {}
                cluster_run["jobId"] = cluster_run_info[0].id
                cluster_run["workDir"] = cluster_def.cluster.work_dir
                cluster_run["mainJobRunId"] = {
                    cluster_run_info[1].name: cluster_run_info[1].id
                }
                if len(cluster_run_info[2]) > 0:
                    cluster_run["otherJobRunIds"] = [
                        {wj.name: wj.id} for wj in cluster_run_info[2]
                    ]
                yamlContent = yaml.dump(cluster_run)
                yamlContent += (
                    "# \u2b50 To stream the logs of the main job run:\n"
                    + f"# \u0024 ads opctl watch {list(cluster_run['mainJobRunId'].values())[0]}"
                )
                print(yamlContent)
                _save_yaml(yamlContent, **kwargs)
            return cluster_run_info
    else:
        if (
            "kind" in p.config
            and p.config["execution"].get("backend", None) != BACKEND_NAME.LOCAL.value
        ):
            p.config["execution"]["backend"] = p.config["kind"]
            return _BackendFactory(p.config).backend.apply()

        if "ocid" in p.config["execution"]:
            resource_to_backend = {
                DataScienceResource.JOB: BACKEND_NAME.JOB,
                DataScienceResource.DATAFLOW: BACKEND_NAME.DATAFLOW,
                DataScienceResource.PIPELINE: BACKEND_NAME.PIPELINE,
            }
            for r, b in resource_to_backend.items():
                if r in p.config["execution"]["ocid"]:
                    p.config["execution"]["backend"] = b.value
        else:
            p.step(ConfigResolver).step(ConfigValidator)
            # spec may have changed during validation step (e.g. defaults filled in)
            # thus command need to be updated since it encodes spec
            p = ConfigResolver(p.config)
            p._resolve_command()
        return _BackendFactory(p.config).backend.run()


def run_diagnostics(config: Dict, **kwargs) -> Dict:
    """
    Run a job given configuration and command line args passed in (kwargs).

    Parameters
    ----------
    config: dict
        dictionary of configurations
    kwargs: dict
        keyword arguments, stores configuration from command line args

    Returns
    -------
    Dict
        dictionary of job id and run id in case of ML Job run, else empty if running locally
    """
    p = ConfigProcessor(config).step(ConfigMerger, **kwargs)
    if config.get("kind") == "distributed":  # TODO: add kind factory

        config = update_config_image(config)
        cluster_def = YamlSpecParser.parse_content(config)

        backend = MLJobDistributedBackend(p.config)

        # Define job first,
        # Then Run
        cluster_run_info = backend.run_diagnostics(
            cluster_info=cluster_def, dry_run=p.config["execution"].get("dry_run")
        )
        if cluster_run_info:
            diagnostics_report_path = os.path.join(
                cluster_def.cluster.work_dir,
                cluster_run_info[0].id,
                "diagnostic_report.html",
            )
            with fsspec.open(
                diagnostics_report_path, "r", **backend.oci_auth
            ) as infile:
                with open(kwargs["output"], "w") as outfile:
                    outfile.write(infile.read())
        return cluster_run_info
    else:
        print("Diagnostics not available for kind: {config.get('kind')}")


def _update_env_vars(config, env_vars: List):
    """
    env_vars: List, should be formatted as [{"name": "OCI__XXX", "value": YYY},]
    """
    # TODO move this to a class which checks the version, kind, type, etc.
    config["spec"]["Runtime"]["spec"]["environmentVariables"].extend(env_vars)
    return config


def init_operator(**kwargs) -> str:
    """
    Initialize the resources for an operator

    Parameters
    ----------
    kwargs: dict
        keyword argument, stores command line args
    Returns
    -------
    folder_path: str
        a path to the folder with all of the resources
    """
    # TODO: confirm that operator slug is in the set of valid operator slugs
    assert kwargs["operator_slug"] == "dask_cluster"

    if kwargs.get("folder_path"):
        kwargs["operator_folder_path"] = kwargs.pop("folder_path")[0]
    else:
        kwargs["operator_folder_path"] = kwargs["operator_slug"]
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    print(f"config check: {p.config}")
    return _BackendFactory(p.config).backend.init_operator()


def delete(**kwargs) -> None:
    """
    Delete a MLJob/DataFlow run.

    Parameters
    ----------
    kwargs: dict
        keyword argument, stores command line args
    Returns
    -------
    None
    """
    kwargs["backend"] = _get_backend_from_ocid(kwargs["ocid"])

    if (
        DataScienceResourceRun.JOB_RUN in kwargs["ocid"]
        or DataScienceResourceRun.DATAFLOW_RUN in kwargs["ocid"]
        or DataScienceResourceRun.PIPELINE_RUN in kwargs["ocid"]
        or DataScienceResourceRun.MODEL_DEPLOYMENT in kwargs["ocid"]
    ):
        kwargs["run_id"] = kwargs.pop("ocid")
    elif (
        DataScienceResource.JOB in kwargs["ocid"]
        or DataScienceResource.DATAFLOW in kwargs["ocid"]
        or DataScienceResource.PIPELINE in kwargs["ocid"]
    ):
        kwargs["id"] = kwargs.pop("ocid")
    else:
        raise ValueError(f"{kwargs['ocid']} is valid or supported.")

    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    return _BackendFactory(p.config).backend.delete()


def cancel(**kwargs) -> None:
    """
    Cancel a MLJob/DataFlow run.

    Parameters
    ----------
    kwargs: dict
        keyword argument, stores command line args
    Returns
    -------
    None
    """
    kwargs["run_id"] = kwargs.pop("ocid")
    if not kwargs.get("backend"):
        kwargs["backend"] = _get_backend_from_run_id(kwargs["run_id"])
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    return _BackendFactory(p.config).backend.cancel()


def watch(**kwargs) -> None:
    """
    Watch a MLJob/DataFlow run.

    Parameters
    ----------
    kwargs: dict
        keyword argument, stores command line args
    Returns
    -------
    None
    """
    kwargs["run_id"] = kwargs.pop("ocid")
    if not kwargs.get("backend"):
        kwargs["backend"] = _get_backend_from_run_id(kwargs["run_id"])
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    return _BackendFactory(p.config).backend.watch()


def activate(**kwargs) -> None:
    """
    Activate a ModelDeployment.

    Parameters
    ----------
    kwargs: dict
        keyword argument, stores command line args

    Returns
    -------
    None
    """
    kwargs["run_id"] = kwargs.pop("ocid")
    if not kwargs.get("backend"):
        kwargs["backend"] = _get_backend_from_run_id(kwargs["run_id"])
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    return _BackendFactory(p.config).backend.activate()


def deactivate(**kwargs) -> None:
    """
    Deactivate a ModelDeployment.

    Parameters
    ----------
    kwargs: dict
        keyword argument, stores command line args

    Returns
    -------
    None
    """
    kwargs["run_id"] = kwargs.pop("ocid")
    if not kwargs.get("backend"):
        kwargs["backend"] = _get_backend_from_run_id(kwargs["run_id"])
    p = ConfigProcessor().step(ConfigMerger, **kwargs)
    return _BackendFactory(p.config).backend.deactivate()


def init_vscode(**kwargs) -> None:
    """
    Create a .devcontainer.json file for local development.

    Parameters
    ----------
    kwargs
        keyword arguments, stores command line args
    Returns
    -------
    None
    """
    p = ConfigMerger({}).process(**kwargs)
    p = ConfigResolver(p.config)
    p._resolve_env_vars()
    p._resolve_mounted_volumes()
    LocalBackend(p.config).init_vscode_container()


def configure() -> None:
    """
    Save default configurations for opctl.

    Parameters
    ----------
    ml_job: bool
        turn on/off saving configurations for ML Job

    Returns
    -------
    None
    """
    folder = os.path.abspath(
        os.path.expanduser(
            click.prompt(
                "Folder to save ADS operators related configurations:",
                default=DEFAULT_ADS_CONFIG_FOLDER,
            )
        )
    )
    os.makedirs(os.path.expanduser(folder), exist_ok=True)

    if os.path.exists(os.path.join(folder, ADS_CONFIG_FILE_NAME)):
        config_parser = read_from_ini(os.path.join(folder, ADS_CONFIG_FILE_NAME))
    else:
        config_parser = configparser.ConfigParser(default_section=None)
        config_parser.optionxform = str

    if "OCI" not in config_parser:
        config_parser["OCI"] = {}
    if "CONDA" not in config_parser:
        config_parser["CONDA"] = {}

    oci_config_path = click.prompt(
        "OCI config path:",
        default=config_parser["OCI"].get("oci_config", DEFAULT_OCI_CONFIG_FILE),
    )
    oci_profile = click.prompt(
        "Default OCI profile:",
        default=config_parser["OCI"].get("oci_profile", DEFAULT_PROFILE),
    )
    oci_config_parser = configparser.ConfigParser()
    oci_config_parser.read(os.path.expanduser(oci_config_path))
    if oci_profile not in oci_config_parser:
        raise ValueError(f"profile {oci_profile} is not found in {oci_config_path}.")
    config_parser["OCI"] = {
        "oci_config": oci_config_path,
        "oci_profile": oci_profile,
    }
    conda_pack_path = click.prompt(
        "Conda pack install folder:",
        default=config_parser["CONDA"].get(
            "conda_pack_folder", DEFAULT_CONDA_PACK_FOLDER
        ),
    )
    config_parser["CONDA"]["conda_pack_folder"] = os.path.abspath(
        os.path.expanduser(conda_pack_path)
    )
    if is_in_notebook_session():
        conda_os_prefix_default = config_parser["CONDA"].get(
            "conda_pack_os_prefix", get_service_pack_prefix()
        )
    else:
        conda_os_prefix_default = config_parser["CONDA"].get("conda_pack_os_prefix", "")

    conda_os_prefix = click.prompt(
        "Object storage Conda Env prefix, in the format oci://<bucket>@<namespace>/<path>",
        default=conda_os_prefix_default,
    )
    config_parser["CONDA"]["conda_pack_os_prefix"] = conda_os_prefix

    with open(os.path.join(os.path.expanduser(folder), ADS_CONFIG_FILE_NAME), "w") as f:
        config_parser.write(f)
    print(f"Configuration saved at {os.path.join(folder, ADS_CONFIG_FILE_NAME)}")

    print("==== Setting configuration for OCI Jobs ====")
    if click.confirm(
        f"Do you want to set up or update OCI Jobs configuration?", default=True
    ):
        required_fields = [
            ("compartment_id", ""),
            ("project_id", ""),
            ("subnet_id", ""),
            ("shape_name", ""),
            ("block_storage_size_in_GBs", ""),
        ]

        optional_fields = [
            ("log_group_id", ""),
            ("log_id", ""),
            ("docker_registry", ""),
            ("conda_pack_os_prefix", "in the format oci://<bucket>@<namespace>/<path>"),
        ]
        _set_service_configurations(
            ADS_JOBS_CONFIG_FILE_NAME,
            required_fields,
            optional_fields,
            folder,
            oci_config_path,
            is_in_notebook_session(),
        )

    print("==== Setting configuration for OCI DataFlow ====")
    if click.confirm(
        f"Do you want to set up or update OCI DataFlow configuration?", default=True
    ):
        required_fields = [
            ("compartment_id", ""),
            ("driver_shape", ""),
            ("executor_shape", ""),
            ("logs_bucket_uri", ""),
            ("script_bucket", "in the format oci://<bucket>@<namespace>/<path>"),
        ]

        optional_fields = [
            ("num_executors", ""),
            ("spark_version", ""),
            ("archive_bucket", "in the format oci://<bucket>@<namespace>/<path>"),
        ]
        _set_service_configurations(
            ADS_DATAFLOW_CONFIG_FILE_NAME,
            required_fields,
            optional_fields,
            folder,
            oci_config_path,
            is_in_notebook_session(),
        )

    print("==== Setting configuration for OCI ML Pipeline ====")
    if click.confirm(
        f"Do you want to set up or update OCI ML Pipeline configuration?", default=True
    ):
        required_fields = [
            ("compartment_id", ""),
            ("project_id", ""),
        ]

        optional_fields = [
            ("log_group_id", ""),
            ("log_id", ""),
        ]
        _set_service_configurations(
            ADS_ML_PIPELINE_CONFIG_FILE_NAME,
            required_fields,
            optional_fields,
            folder,
            oci_config_path,
            is_in_notebook_session(),
        )

    print("==== Setting configuration for local backend ====")
    if click.confirm(
        f"Do you want to set up or update local backend configuration?", default=True
    ):
        required_fields = [
            ("max_parallel_containers", str(min(os.cpu_count(), 4))),
            ("pipeline_status_poll_interval_seconds", str(5)),
        ]

        optional_fields = []

        _set_service_configurations(
            ADS_LOCAL_BACKEND_CONFIG_FILE_NAME,
            required_fields,
            optional_fields,
            folder,
            oci_config_path,
            is_in_notebook_session(),
        )


def _set_service_configurations(
    config_file_name,
    required_fields,
    optional_fields,
    config_folder,
    oci_config_path=None,
    rp=False,
):
    if os.path.exists(os.path.join(config_folder, config_file_name)):
        infra_parser = read_from_ini(os.path.join(config_folder, config_file_name))
    else:
        infra_parser = configparser.ConfigParser(default_section=None)
        infra_parser.optionxform = str

    def prompt_for_values(infra_parser, profile):
        for field, msg in required_fields:
            prompt = f"Specify {field}" if len(msg) == 0 else f"Specify {field}, {msg}"
            infra_parser[profile][field] = click.prompt(
                prompt,
                default=infra_parser[profile].get(field, None),
            )
        for field, msg in optional_fields:
            prompt = f"Specify {field}" if len(msg) == 0 else f"Specify {field}, {msg}"
            ans = click.prompt(
                prompt,
                default=infra_parser[profile].get(field, ""),
            )
            if len(ans) > 0:
                infra_parser[profile][field] = ans

    if rp:
        if "RESOURCE_PRINCIPAL" not in infra_parser:
            with AuthContext():
                ads.set_auth(auth=AuthType.RESOURCE_PRINCIPAL)
                notebook_session = DSCNotebookSession.from_ocid(
                    os.environ["NB_SESSION_OCID"]
                )
            default_values = {
                "compartment_id": notebook_session.compartment_id,
                "project_id": notebook_session.project_id,
                "subnet_id": notebook_session.notebook_session_configuration_details.subnet_id,
                "block_storage_size_in_GBs": notebook_session.notebook_session_configuration_details.block_storage_size_in_gbs,
                "shape_name": notebook_session.notebook_session_configuration_details.shape,
                "driver_shape": notebook_session.notebook_session_configuration_details.shape,
                "executor_shape": notebook_session.notebook_session_configuration_details.shape,
            }
            infra_parser["RESOURCE_PRINCIPAL"] = {
                k: v
                for k, v in default_values.items()
                if (
                    k in [field[0] for field in required_fields]
                    or k in [field[0] for field in optional_fields]
                )
            }
        if click.confirm(
            "Do you want to set up or update configuration when using resource principal?",
            default=True,
        ):
            prompt_for_values(infra_parser, "RESOURCE_PRINCIPAL")

    if os.path.exists(os.path.expanduser(oci_config_path)):
        parser = read_from_ini(os.path.abspath(os.path.expanduser(oci_config_path)))
        for oci_profile in parser:
            if oci_profile:
                if click.confirm(
                    f"Do you want to set up or update for profile {oci_profile}?"
                ):
                    if oci_profile not in infra_parser:
                        infra_parser[oci_profile] = {}
                    prompt_for_values(infra_parser, oci_profile)

    with open(os.path.join(config_folder, config_file_name), "w") as f:
        infra_parser.write(f)
    print(f"Configuration saved at {os.path.join(config_folder, config_file_name)}")


def _get_backend_from_ocid(ocid: str) -> str:
    for value in DataScienceResource.values():
        if value in ocid:
            return DATA_SCIENCE_RESOURCE_BACKEND_MAP[value]
    raise ValueError("Must provide a resource OCID.")


def _get_backend_from_run_id(ocid: str) -> str:
    for value in DataScienceResourceRun.values():
        if value in ocid:
            return DATA_SCIENCE_RESOURCE_RUN_BACKEND_MAP[value]
    raise ValueError("Must provide a resource run OCID.")



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import importlib
import multiprocessing
import os
import uuid
import psutil
from enum import Enum, auto
from time import time, sleep

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import logging
from ads.common import logger
from ads.common import utils
from ads.common.data import ADSData
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.hpo._imports import try_import
from ads.hpo.ads_search_space import get_model2searchspace
from ads.hpo.distributions import *
from ads.hpo.objective import _Objective
from ads.hpo.stopping_criterion import NTrials, ScoreValue, TimeBudget
from ads.hpo.utils import _num_samples, _safe_indexing, _update_space_name
from ads.hpo.validation import (
    assert_is_estimator,
    assert_model_is_supported,
    assert_strategy_valid,
    assert_tuner_is_fitted,
    validate_fit_params,
    validate_pipeline,
    validate_search_space,
    validate_params_for_plot,
)


with try_import() as _imports:
    from sklearn.base import BaseEstimator, clone, is_classifier
    from sklearn.model_selection import BaseCrossValidator  # NOQA
    from sklearn.model_selection import check_cv, cross_validate
    from sklearn.pipeline import Pipeline, make_pipeline
    from sklearn.utils import check_random_state
    from sklearn.exceptions import NotFittedError

    try:
        from sklearn.metrics import check_scoring
    except:
        from sklearn.metrics.scorer import check_scoring


from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Union  # NOQA


class State(Enum):
    INITIATED = auto()
    RUNNING = auto()
    HALTED = auto()
    TERMINATED = auto()
    COMPLETED = auto()


class InvalidStateTransition(Exception):
    """
    `Invalid State Transition` is raised when an invalid transition request is made, such as calling
    halt without a running process.
    """

    pass


class ExitCriterionError(Exception):
    """
    `ExitCriterionError` is raised when an attempt is made to check exit status for a different exit
    type than the tuner was initialized with. For example, if an HPO study has an exit criteria based
    on the number of trials and a request is made for the time remaining, which is a different exit
    criterion, an exception is raised.
    """

    pass


class DuplicatedStudyError(Exception):
    """
    `DuplicatedStudyError` is raised when a new tuner process is created with a study name that
    already exists in storage.
    """


class NoRestartError(Exception):
    """
    `NoRestartError` is raised when an attempt is made to check how many seconds have transpired since
    the HPO process was last resumed from a halt. This can happen if the process has been terminated
    or it was never halted and then resumed to begin with.
    """

    pass


class DataScienceObjective:
    """This class is to replace the previous lambda function to solve the problem that python does not allow pickle local function/lambda function."""

    def __init__(self, objective, X_res, y_res):
        self.objective = objective
        self.X_res = X_res
        self.y_res = y_res

    def __call__(self, trial):
        return self.objective(self.X_res, self.y_res, trial)


class ADSTuner(BaseEstimator):
    """
    Hyperparameter search with cross-validation.
    """

    _required_parameters = ["model"]

    @property
    def sklearn_steps(self):
        """
        Returns
        -------
            int
                Search space which corresponds to the best candidate parameter setting.
        """
        return _update_space_name(self.best_params, step_name=self._step_name)

    @property
    def best_index(self):
        """
        Returns
        -------
            int
                Index which corresponds to the best candidate parameter setting.
        """
        return self.trials["value"].idxmax()

    @property
    def best_params(self):
        """
        Returns
        -------
            Dict[str, Any]
                Parameters of the best trial.
        """
        self._check_is_fitted()
        return self._remove_step_name(self._study.best_params)

    @property
    def best_score(self):
        """
        Returns
        -------
            float
                Mean cross-validated score of the best estimator.
        """
        self._check_is_fitted()
        return self._study.best_value

    @property
    def score_remaining(self):
        """
        Returns
        -------
            float
                The difference between the best score and the optimal score.

        Raises
        ------
            :class:`ExitCriterionError`
                Error is raised if there is no score-based criteria for tuning.
        """
        if self._optimal_score is None:
            raise ExitCriterionError(
                "Tuner does not have a score-based exit condition."
            )
        else:
            return self._optimal_score - self.best_score

    @property
    def scoring_name(self):
        """
        Returns
        -------
            str
                Scoring name.
        """
        return self._extract_scoring_name()

    @property
    def n_trials(self):
        """
        Returns
        -------
            int
                Number of completed trials. Alias for `trial_count`.
        """
        self._check_is_fitted()
        return len(self.trials)

    # Alias for n_trials
    trial_count = n_trials

    @property
    def trials_remaining(self):
        """
        Returns
        -------
            int
                The number of trials remaining in the budget.

        Raises
        ------
            :class:`ExitCriterionError`
                Raised if the current tuner does not include a trials-based exit
                condition.
        """
        if self._n_trials is None:
            raise ExitCriterionError(
                "This tuner does not include a trials-based exit condition"
            )
        return self._n_trials - self.n_trials + self._previous_trial_count

    @property
    def trials(self):
        """
        Returns
        -------
            :class:`pandas.DataFrame`
                Trial data up to this point.
        """
        if self.is_halted():
            if self._trial_dataframe is None:
                return pd.DataFrame()
            return self._trial_dataframe
        trials_dataframe = self._study.trials_dataframe().copy()
        return trials_dataframe

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def __init__(
        self,
        model,  # type: Union[BaseEstimator, Pipeline]
        strategy="perfunctory",  # type: Union[str, Mapping[str, optuna.distributions.BaseDistribution]]
        scoring=None,  # type: Optional[Union[Callable[..., float], str]]
        cv=5,  # type: Optional[int]
        study_name=None,  # type: Optional[str]
        storage=None,  # type: Optional[str]
        load_if_exists=True,  # type: Optional[bool]
        random_state=None,  # type: Optional[int]
        loglevel=logging.INFO,  # type: Optional[int]
        n_jobs=1,  # type: Optional[int]
        X=None,  # type: Union[List[List[float]], np.ndarray, pd.DataFrame, spmatrix, ADSData]
        y=None,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
    ):
        # type: (...) -> None
        """
        Returns a hyperparameter tuning object

        Parameters
        ----------
            model:
                Object to use to fit the data. This is assumed to implement the
                scikit-learn estimator or pipeline interface.
            strategy:
                ``perfunctory``, ``detailed`` or a dictionary/mapping of hyperparameter
                and its distribution . If obj:`perfunctory`, picks a few
                relatively more important hyperparmeters to tune . If obj:`detailed`,
                extends to a larger search space. If obj:dict, user defined search
                space: Dictionary where keys are hyperparameters and values are distributions.
                Distributions are assumed to implement the ads distribution interface.
            scoring: Optional[Union[Callable[..., float], str]]
                String or callable to evaluate the predictions on the validation data.
                If :obj:`None`, ``score`` on the estimator is used.
            cv: int
                Integer to specify the number of folds in a CV splitter.
                If :obj:`estimator` is a classifier and :obj:`y` is
                either binary or multiclass,
                ``sklearn.model_selection.StratifiedKFold`` is used. otherwise,
                ``sklearn.model_selection.KFold`` is used.
            study_name: str,
                Name of the current experiment for the ADSTuner object. One ADSTuner
                object can only be attached to one study_name.
            storage:
                Database URL. (e.g. sqlite:///example.db). Default to sqlite:////tmp/hpo_*.db.
            load_if_exists:
                Flag to control the behavior to handle a conflict of study names.
                In the case where a study named ``study_name`` already exists in the ``storage``,
                a :class:`DuplicatedStudyError` is raised if ``load_if_exists`` is
                set to :obj:`False`.
                Otherwise, the existing one is returned.
            random_state:
                Seed of the pseudo random number generator. If int, this is the
                seed used by the random number generator. If :obj:`None`, the global random state from
                ``numpy.random`` is used.
            loglevel:
                loglevel. can be logging.NOTSET, logging.INFO, logging.DEBUG, logging.WARNING
            n_jobs: int
                Number of parallel jobs. :obj:`-1` means using all processors.
            X: TwoDimArrayLikeType, Union[List[List[float]], np.ndarray,
            pd.DataFrame, spmatrix, ADSData]
                Training data.
            y: Union[OneDimArrayLikeType, TwoDimArrayLikeType], optional
            OneDimArrayLikeType: Union[List[float], np.ndarray, pd.Series]
            TwoDimArrayLikeType: Union[List[List[float]], np.ndarray, pd.DataFrame, spmatrix, ADSData]
                Target.

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.svm import SVC

            tuner = ADSTuner(
                            SVC(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )

            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
        """
        _imports.check()
        self._n_jobs = n_jobs
        assert (
            cv > 1
        ), "k-fold cross-validation requires at least one train/test split by setting cv=2 or more"
        self.cv = cv
        self._error_score = np.nan
        self.model = model
        self._check_pipeline()
        self._step_name = None
        self._extract_estimator()
        self.strategy = None
        self._param_distributions = None
        self._check_strategy(strategy)
        self.strategy = strategy
        self._param_distributions = self._get_param_distributions(self.strategy)
        self._enable_pruning = hasattr(self.model, "partial_fit")
        self._max_iter = 100
        self.__random_state = random_state  # to be used in export_trials
        # this calls the randomstate.setter which turns self.random_state into a np.random.RandomState instance
        # make it hard to be serialized.
        self.random_state = check_random_state(random_state)

        self._return_train_score = False
        self.scoring = scoring
        self._subsample = 1.0
        self.loglevel = loglevel
        self._trial_dataframe = None
        self._status = State.INITIATED
        self.study_name = (
            study_name if study_name is not None else "hpo_" + str(uuid.uuid4())
        )
        self.storage = (
            "sqlite:////tmp/hpo_" + str(uuid.uuid4()) + ".db"
            if storage is None
            else storage
        )
        self.oci_client = None

        seed = np.random.randint(0, np.iinfo("int32").max)

        self.sampler = optuna.samplers.TPESampler(seed=seed)
        self.median_pruner = self._pruner(
            class_name="median_pruner",
            n_startup_trials=5,
            n_warmup_steps=1,
            interval_steps=1,
        )
        self.load_if_exists = load_if_exists
        try:
            self._study = optuna.study.create_study(
                study_name=self.study_name,
                direction="maximize",
                pruner=self.median_pruner,
                sampler=self.sampler,
                storage=self.storage,
                load_if_exists=self.load_if_exists,
            )
        except optuna.exceptions.DuplicatedStudyError as e:
            if self.load_if_exists:
                logger.info(
                    "Using an existing study with name '{}' instead of "
                    "creating a new one.".format(self.study_name)
                )
            else:
                raise DuplicatedStudyError(
                    f"The study_name `{self.study_name}` exists in the {self.storage}. Either set load_if_exists=True, or use a new study_name."
                )
        self._init_data(X, y)

    def search_space(self, strategy=None, overwrite=False):
        """
        Returns the search space. If strategy is not passed in, return the existing search
        space. When strategy is passed in, overwrite the existing search space if overwrite
        is set True, otherwise, only update the existing search space.

        Parameters
        ----------
        strategy: Union[str, dict], optional
            ``perfunctory``, ``detailed`` or a dictionary/mapping of the hyperparameters
            and their distributions. If obj:`perfunctory`, picks a few relatively
            more important hyperparmeters to tune . If obj:`detailed`, extends to a
            larger search space. If obj:dict, user defined search space: Dictionary
            where keys are parameters and values are distributions. Distributions are
            assumed to implement the ads distribution interface.
        overwrite: bool, optional
            Ignored when strategy is None. Otherwise, search space is overwritten if overwrite
            is set True and updated if it is False.

        Returns
        -------
            dict
                A mapping of the hyperparameters and their distributions.

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
            tuner.search_space()
        """
        assert hasattr(
            self, "_param_distributions"
        ), "Call <code>ADSTuner</code> first."
        if not strategy:
            return self._remove_step_name(self._param_distributions)
        self._check_strategy(strategy)
        self.strategy = strategy
        if overwrite:
            self._param_distributions = self._get_param_distributions(self.strategy)
        else:
            self._param_distributions.update(
                self._get_param_distributions(self.strategy)
            )
        return self._remove_step_name(self._param_distributions)

    @staticmethod
    def _remove_step_name(param_distributions):
        search_space = {}
        for param, distributions in param_distributions.items():
            if "__" in param:
                param = param.split("__")[1]
            search_space[param] = distributions
        return search_space

    def _check_pipeline(self):
        self.model = validate_pipeline(self.model)

    def _get_internal_param_distributions(self, strategy):
        if isinstance(self.model, Pipeline):
            for step_name, step in self.model.steps:
                if step.__class__ in get_model2searchspace().keys():
                    self._step_name = step_name
                    param_distributions = get_model2searchspace()[step.__class__](
                        strategy
                    ).suggest_space(step_name=step_name)
            if len(param_distributions) == 0:
                logger.warning("Nothing to tune.")
        else:
            assert_model_is_supported(self.model)
            param_distributions = get_model2searchspace()[self.model.__class__](
                strategy
            ).suggest_space()
            self._check_search_space(param_distributions)
        return param_distributions

    def _get_param_distributions(self, strategy):
        if isinstance(strategy, str):
            param_distributions = self._get_internal_param_distributions(strategy)
        if isinstance(strategy, dict):
            param_distributions = _update_space_name(
                strategy, step_name=self._step_name
            )
            self._check_search_space(param_distributions)
        return param_distributions

    def _check_search_space(self, param_distributions):

        validate_search_space(self.model.get_params().keys(), param_distributions)

    def _check_is_fitted(self):
        assert_tuner_is_fitted(self)

    def _check_strategy(self, strategy):
        assert_strategy_valid(self._param_distributions, strategy, self.strategy)

    def _add_halt_time(self):
        """Adds a new start time window to the start/stop log. This happens in two cases: when the tuning process
        has commenced and when it resumes following a halt
        """
        self._time_log.append(dict(halt=time(), resume=None))

    def _add_resume_time(self):
        """Adds a new stopping time to the last window in the time log. This happens when the HPO process is
        halted or terminated.
        """
        if len(self._time_log) > 0:
            entry = self._time_log.pop()
        if entry["resume"] is not None:
            raise Exception("Cannot close a time window without an opening time.")
        self._time_log.append(dict(halt=entry["halt"], resume=time()))

    def tune(
        self,
        X=None,  # type: TwoDimArrayLikeType
        y=None,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        exit_criterion=[],  # type: Optional[list]
        loglevel=None,  # type: Optional[int]
        synchronous=False,  # type: Optional[boolean]
    ):
        """
        Run hypyerparameter tuning until one of the <code>exit_criterion</code>
        is met. The default is to run 50 trials.

        Parameters
        ----------
        X: TwoDimArrayLikeType, Union[List[List[float]], np.ndarray, pd.DataFrame, spmatrix, ADSData]

            Training data.
        y: Union[OneDimArrayLikeType, TwoDimArrayLikeType], optional
        OneDimArrayLikeType: Union[List[float], np.ndarray, pd.Series]
        TwoDimArrayLikeType: Union[List[List[float]], np.ndarray, pd.DataFrame, spmatrix, ADSData]

            Target.
        exit_criterion: list, optional
            A list of ads stopping criterion. Can be `ScoreValue()`, `NTrials()`, `TimeBudget()`.
            For example, [ScoreValue(0.96), NTrials(40), TimeBudget(10)]. It will exit when any of the
            stopping criterion is satisfied in the `exit_criterion` list.
            By default, the run will stop after 50 trials.
        loglevel: int, optional
            Log level.
        synchronous: boolean, optional
            Tune synchronously or not. Defaults to `False`

        Returns
        -------
            None
                Nothing

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.svm import SVC

            tuner = ADSTuner(
                            SVC(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
        """

        # Get previous trial count to ensure proper counting.
        try:
            self._previous_trial_count = self.trial_count
        except NotFittedError:
            self._previous_trial_count = 0
        except Exception as e:
            _logger.error(f"Error retrieving previous trial count: {e}")
            raise

        self._init_data(X, y)
        if self.X is None:
            raise ValueError(
                "Need to either pass the data to `X` and `y` in `tune()`, or to `ADSTuner`."
            )
        if self.is_running():
            raise InvalidStateTransition(
                "Running process found. Do you need to call terminate() to stop before calling tune()?"
            )
        if self.is_halted():
            raise InvalidStateTransition(
                "Halted process found. You need to call resume()."
            )
        # handle ADSData

        # Initialize time log for every new call to tune(). Set shared global time values
        self._global_start = multiprocessing.Value("d", 0.0)
        self._global_stop = multiprocessing.Value("d", 0.0)
        self._time_log = []

        self._tune(
            X=self.X,
            y=self.y,
            exit_criterion=exit_criterion,
            loglevel=loglevel,
            synchronous=synchronous,
        )

        # Tune cannot exit before the clock starts in the subprocess.
        while self._global_start.value == 0.0:
            sleep(0.01)

    def _init_data(self, X, y):
        if X is not None:
            if isinstance(X, ADSData):
                self.y = X.y
                self.X = X.X
            else:
                self.X = X
                self.y = y

    def halt(self):
        """
        Halt the current running tuning process.

        Returns
        -------
            None
                Nothing

        Raises
        ------
            `InvalidStateTransition` if no running process is found

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
            tuner.halt()
        """
        if hasattr(self, "_tune_process") and self._status == State.RUNNING:
            self._trial_dataframe = self._study.trials_dataframe().copy()
            psutil.Process(self._tune_process.pid).suspend()
            self._status = State.HALTED
            self._add_halt_time()
        else:
            raise InvalidStateTransition(
                "No running process found. Do you need to call tune()?"
            )

    def resume(self):
        """
        Resume the current halted tuning process.

        Returns
        -------
            None
                Nothing

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
            tuner.halt()
            tuner.resume()
        """
        if self.is_halted():
            psutil.Process(self._tune_process.pid).resume()
            self._add_resume_time()
            self._status = State.RUNNING
        else:
            raise InvalidStateTransition("No paused process found.")

    def wait(self):
        """
        Wait for the current tuning process to finish running.

        Returns
        -------
            None
                Nothing

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
            tuner.wait()
        """
        if self.is_running():
            self._tune_process.join()
            self._status = State.COMPLETED
        else:
            raise InvalidStateTransition("No running process.")

    def terminate(self):
        """
        Terminate the current tuning process.

        Returns
        -------
            None
                Nothing

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
            tuner.terminate()
        """
        if self.is_running():
            self._tune_process.terminate()
            self._tune_process.join()
            self._status = State.TERMINATED
            # self._add_terminate_time()
            self._update_failed_trial_state()
        else:
            raise RuntimeError("No running process found. Do you need to call tune()?")

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def _update_failed_trial_state(self):
        from optuna.trial import TrialState

        for trial in self._study.trials:
            if trial.state == TrialState.RUNNING:
                self._study._storage.set_trial_state(
                    trial._trial_id, optuna.structs.TrialState.FAIL
                )

    @property
    def time_remaining(self):
        """Returns the number of seconds remaining in the study

        Returns
        -------
            int: Number of seconds remaining in the budget. 0 if complete/terminated

        Raises
        ------
            :class:`ExitCriterionError`
                Error is raised if time has not been included in the budget.
        """
        if self._time_budget is None:
            raise ExitCriterionError(
                "This tuner does not include a time-based exit condition"
            )
        elif self.is_completed() or self.is_terminated():
            return 0
        return max(self._time_budget - self.time_elapsed, 0)

    @property
    def time_since_resume(self):
        """Return the seconds since the process has been resumed from a halt.

        Returns
        -------
            int: the number of seconds since the process was last resumed

        Raises
        ------
            `NoRestartError` is the process has not been resumed

        """
        if len(self._time_log) > 0:
            last_time_resumed = self._time_log[-1].get("resume")
        else:
            raise Exception("Time log should not be empty")

        if self.is_running():
            if last_time_resumed is not None:
                return time() - last_time_resumed
            else:
                raise NoRestartError("The process has not been resumed")
        elif self.is_halted():
            return 0  # if halted, the amount of time since restarted from a halt is 0
        elif self.is_terminated():
            raise NoRestartError("The process has been terminated")

    @property
    def time_elapsed(self):
        """Return the time in seconds that the HPO process has been searching

        Returns
        -------
            int: The number of seconds the HPO process has been searching
        """
        time_in_halted_state = 0.0

        # Add up all the halted durations, i.e. the time spent between halt and resume
        for entry in self._time_log:
            halt_time = entry.get("halt")
            resume_time = entry.get("resume")

            if resume_time is None:
                # halted state.
                # elapsed = halt time - global start - time halted
                elapsed = halt_time - self._global_start.value - time_in_halted_state
                return elapsed

            else:
                # running/completed/terminated state,
                time_in_halted_state += resume_time - halt_time

        # If the loop ends all halts were resumed. If self._global_stop != 0 that means the
        # process has exited.
        if self._global_stop.value != 0:
            global_time = self._global_stop.value - self._global_start.value
        else:
            global_time = time() - self._global_start.value

        elapsed = global_time - time_in_halted_state
        return elapsed

    def best_scores(self, n: int = 5, reverse: bool = True):
        """Return the best scores from the study

        Parameters
        ----------
        n: int
            The maximum number of results to show. Defaults to 5. If `None` or
            negative return all.
        reverse: bool
            Whether to reverse the sort order so results are in descending order.
            Defaults to `True`

        Returns
        -------
        list[float or int]
            List of the best scores

        Raises
        ------
        `ValueError` if there are no trials
        """
        if len(self.trials) < 1:
            raise ValueError("No score data to show")
        else:
            scores = self.trials.value
            scores = scores[scores.notnull()]
            if scores is None:
                raise ValueError(
                    f"No score data despite valid trial data. Trial data length: {len(self.trials)}"
                )
            if not isinstance(n, int) or n <= 0:
                return sorted(scores, reverse=reverse)
            else:
                return sorted(scores, reverse=reverse)[:n]

    def get_status(self):
        """
        return the status of the current tuning process.

        Alias for the property `status`.

        Returns
        -------
            :class:`Status`
                The status of the process

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)])
            tuner.get_status()
        """
        return self.status

    def is_running(self):
        """
        Returns
        -------
            bool
                `True` if the :class:`ADSTuner` instance is running; `False` otherwise.
        """
        return self.status == State.RUNNING

    def is_halted(self):
        """
        Returns
        -------
            bool
                `True` if the :class:`ADSTuner` instance is halted; `False` otherwise.
        """
        return self.status == State.HALTED

    def is_terminated(self):
        """
        Returns
        -------
            bool
                `True` if the :class:`ADSTuner` instance has been terminated; `False` otherwise.
        """
        return self.status == State.TERMINATED

    def is_completed(self):
        """
        Returns
        -------
            bool
                `True` if the :class:`ADSTuner` instance has completed; `False` otherwise.
        """
        return self.status == State.COMPLETED

    def _is_tuning_started(self):
        """
        Returns
        -------
            bool
                `True` if the :class:`ADSTuner` instance has been started (for example, halted or
                running); `False` otherwise.
        """
        return self.status == State.HALTED or self.status == State.RUNNING

    def _is_tuning_finished(self):
        """
        Returns
        -------
            bool
                `True` if the :class:`ADSTuner` instance is finished running (i.e. completed
                or terminated); `False` otherwise.
        """
        return self.status == State.COMPLETED or self.status == State.TERMINATED

    @property
    def status(self):
        """
        Returns
        -------
            :class:`Status`
                The status of the current tuning process.
        """
        if (
            self._status == State.HALTED
            or self._status == State.TERMINATED
            or self._status == State.INITIATED
        ):
            return self._status
        elif hasattr(self, "_tune_process") and self._tune_process.is_alive():
            return State.RUNNING
        else:
            return State.COMPLETED
        return self._status

    def _extract_exit_criterion(self, exit_criterion):
        # handle the exit criterion
        self._time_budget = None
        self._n_trials = None
        self.exit_criterion = []
        self._optimal_score = None
        if exit_criterion is None or len(exit_criterion) == 0:
            self._n_trials = 50
        for i, criteria in enumerate(exit_criterion):
            if isinstance(criteria, TimeBudget):
                self._time_budget = criteria()
            elif isinstance(criteria, NTrials):
                self._n_trials = criteria()
            elif isinstance(criteria, ScoreValue):
                self._optimal_score = criteria.score
                self.exit_criterion.append(criteria)
            else:
                raise NotImplementedError(
                    "``{}`` is not supported!".format(criteria.__class__.__name__)
                )

    def _extract_estimator(self):
        if isinstance(self.model, Pipeline):  # Pipeline
            for step_name, step in self.model.steps:
                if self._is_estimator(step):
                    self._step_name = step_name
                    self.estimator = step

        else:
            self.estimator = self.model
            assert_is_estimator(self.estimator)
            # assert _check_estimator(self.estimator), "Estimator must implement fit"

    def _extract_scoring_name(self):
        if isinstance(self.scoring, str):
            return self.scoring
        if self._scorer.__class__.__name__ != "function":
            return (
                self._scorer
                if isinstance(self._scorer, str)
                else str(self._scorer).split("(")[1].split(")")[0]
            )
        else:
            if is_classifier(self.model):
                return "mean accuracy"
            else:
                return "r2"

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def _set_logger(self, loglevel, class_name):
        if loglevel is not None:
            self.loglevel = loglevel
        if class_name == "optuna":
            optuna.logging.set_verbosity(self.loglevel)
        else:
            raise NotImplementedError("{} is not supported.".format(class_name))

    def _set_sample_indices(self, X, random_state):
        max_samples = self._subsample
        n_samples = _num_samples(X)
        self._sample_indices = np.arange(n_samples)

        if isinstance(max_samples, float):
            max_samples = int(max_samples * n_samples)

        if max_samples < n_samples:
            self._sample_indices = random_state.choice(
                self._sample_indices, max_samples, replace=False
            )

            self._sample_indices.sort()

    def _get_fit_params_res(self, X):
        fit_params = {}
        fit_params_res = fit_params

        if fit_params_res is not None:
            fit_params_res = validate_fit_params(X, fit_params, self._sample_indices)
        return fit_params_res

    def _can_tune(self):
        assert hasattr(self, "model"), "Call <code>ADSTuner</code> first."
        if self._param_distributions == {}:
            logger.warning("Nothing to tune.")

        if self._param_distributions is None:
            raise NotImplementedError(
                "There was no model specified or the model is not supported."
            )

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def _tune(
        self,
        X,  # type: TwoDimArrayLikeType
        y,  # type: Optional[Union[OneDimArrayLikeType, TwoDimArrayLikeType]]
        exit_criterion=[],  # type: Optional[list]
        loglevel=None,  # type: Optional[int]
        synchronous=False,  # type: Optional[boolean]
    ):
        # type: (...) -> tuple
        """
        Tune with all sets of parameters.
        """
        self._can_tune()
        self._set_logger(loglevel=loglevel, class_name="optuna")
        self._extract_exit_criterion(exit_criterion)
        self._extract_estimator()
        random_state = self.random_state
        old_level = logger.getEffectiveLevel()
        logger.setLevel(self.loglevel)
        if not synchronous:
            optuna.logging.set_verbosity(optuna.logging.ERROR)
            logger.setLevel(logging.ERROR)

        self._set_sample_indices(X, random_state)
        X_res = _safe_indexing(X, self._sample_indices)
        y_res = _safe_indexing(y, self._sample_indices)
        groups_res = _safe_indexing(None, self._sample_indices)
        fit_params_res = self._get_fit_params_res(X)

        classifier = is_classifier(self.model)
        cv = check_cv(self.cv, y_res, classifier=classifier)
        self._n_splits = cv.get_n_splits(X_res, y_res, groups=groups_res)

        # scoring
        self._scorer = check_scoring(self.estimator, scoring=self.scoring)

        self._study = optuna.study.create_study(
            study_name=self.study_name,
            direction="maximize",
            pruner=self.median_pruner,
            sampler=self.sampler,
            storage=self.storage,
            load_if_exists=self.load_if_exists,
        )
        objective = _Objective(
            self.model,
            self._param_distributions,
            cv,
            self._enable_pruning,
            self._error_score,
            fit_params_res,
            groups_res,
            self._max_iter,
            self._return_train_score,
            self._scorer,
            self.scoring_name,
            self._step_name,
        )

        if synchronous:
            logger.info(
                "Optimizing hyperparameters using {} "
                "samples...".format(_num_samples(self._sample_indices))
            )

        self._tune_process = multiprocessing.Process(
            target=ADSTuner.optimizer,
            args=(
                self.study_name,
                self.median_pruner,
                self.sampler,
                self.storage,
                self.load_if_exists,
                DataScienceObjective(objective, X_res, y_res),
                self._global_start,
                self._global_stop,
            ),
            kwargs=dict(
                n_jobs=self._n_jobs,
                n_trials=self._n_trials,
                timeout=self._time_budget,
                show_progress_bar=False,
                callbacks=self.exit_criterion,
                gc_after_trial=False,
            ),
        )

        self._tune_process.start()
        self._status = State.RUNNING

        if synchronous:
            self._tune_process.join()
            logger.info("Finished hyperparemeter search!")
            self._status = State.COMPLETED

        logger.setLevel(old_level)

    @staticmethod
    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def optimizer(
        study_name,
        pruner,
        sampler,
        storage,
        load_if_exists,
        objective_func,
        global_start,
        global_stop,
        **kwargs,
    ):
        """
        Static method for running ADSTuner tuning process

        Parameters
        ----------
            study_name: str
                The name of the study.
            pruner
                The pruning method for pruning trials.
            sampler
                The sampling method used for tuning.
            storage: str
                Storage endpoint.
            load_if_exists: bool
                Load existing study if it exists.
            objective_func
                The objective function to be maximized.
            global_start: :class:`multiprocesing.Value`
                The global start time.
            global_stop: :class:`multiprocessing.Value`
                The global stop time.
            kwargs: dict
                Keyword/value pairs passed into the optimize process


        Raises
        ------
            :class:`Exception`
                Raised for any exceptions thrown by the underlying optimization process

        Returns
        -------
            None
                Nothing

        """
        import traceback

        study = optuna.study.create_study(
            study_name=study_name,
            direction="maximize",
            pruner=pruner,
            sampler=sampler,
            storage=storage,
            load_if_exists=load_if_exists,
        )
        try:
            global_start.value = time()
            study.optimize(objective_func, **kwargs)
            global_stop.value = time()
        except Exception as e:
            traceback.print_exc()
            raise e

    @staticmethod
    def _is_estimator(step):
        return hasattr(step, "fit") and (
            not hasattr(step, "transform")
            or hasattr(step, "predict")
            or hasattr(step, "fit_predict")
        )

    @staticmethod
    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def _pruner(class_name, **kwargs):
        if class_name == "median_pruner":
            return optuna.pruners.MedianPruner(**kwargs)
        else:
            raise NotImplementedError("{} is not supported.".format(class_name))

    def trials_export(
        self, file_uri, metadata=None, script_dict={"model": None, "scoring": None}
    ):
        """Export the meta data as well as files needed to reconstruct the ADSTuner object to the object storage.
        Data is not stored. To resume the same ADSTuner object from object storage and continue tuning from previous trials,
        you have to provide the dataset.

        Parameters
        ----------
            file_uri: str
                Object storage path, 'oci://bucketname@namespace/filepath/on/objectstorage'. For example,
                `oci://test_bucket@ociodsccust/tuner/test.zip`
            metadata: str, optional
                User defined metadata
            script_dict: dict, optional
                Script paths for model and scoring. This is only recommended for unsupported
                models and user-defined scoring functions. You can store the model and scoring
                function in a dictionary with keys `model` and `scoring` and the respective
                paths as values. The model and scoring scripts must import necessary libraries
                for the script to run. The ``model`` and ``scoring`` variables must be set to
                your model and scoring function.

        Returns
        -------
            None
                Nothing

        Example::

            # Print out a list of supported models
            from ads.hpo.ads_search_space import model_list
            print(model_list)

            # Example scoring dictionary
            {'model':'/home/datascience/advanced-ds/notebooks/scratch/ADSTunerV2/mymodel.py',
            'scoring':'/home/datascience/advanced-ds/notebooks/scratch/ADSTunerV2/customized_scoring.py'}

        Example::

            from ads.hpo.stopping_criterion import *
            from ads.hpo.search_cv import ADSTuner
            from sklearn.datasets import load_iris
            from sklearn.linear_model import SGDClassifier

            tuner = ADSTuner(
                            SGDClassifier(),
                            strategy='detailed',
                            scoring='f1_weighted',
                            random_state=42
                        )
            tuner.search_space({'max_iter': 100})
            X, y = load_iris(return_X_y=True)
            tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)], synchronous=True)
            tuner.trials_export('oci://<bucket_name>@<namespace>/tuner/test.zip')
        """
        # oci://bucketname@namespace/filename
        from ads.hpo.tuner_artifact import UploadTunerArtifact

        assert self._is_tuning_finished()
        assert script_dict.keys() <= set(
            ["model", "scoring"]
        ), "script_dict keys can only be model and scoring."

        UploadTunerArtifact(self, file_uri, metadata).upload(script_dict)

    @classmethod
    def trials_import(cls, file_uri, delete_zip_file=True, target_file_path=None):
        """Import the database file from the object storage

        Parameters
        ----------
        file_uri: str
            'oci://bucketname@namespace/filepath/on/objectstorage'
            Example: 'oci://<bucket_name>@<namespace>/tuner/test.zip'
        delete_zip_file: bool, defaults to True, optional
            Whether delete the zip file afterwards.
        target_file_path: str, optional
            The path where the zip file will be saved. For example, '/home/datascience/myfile.zip'.

        Returns
        -------
        :class:`ADSTuner`
            ADSTuner object

        Examples
        --------
        >>> from ads.hpo.stopping_criterion import *
        >>> from ads.hpo.search_cv import ADSTuner
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.linear_model import SGDClassifier
        >>> X, y = load_iris(return_X_y=True)
        >>> tuner = ADSTuner.trials_import('oci://<bucket_name>@<namespace>/tuner/test.zip')
        >>> tuner.tune(X=X, y=y, exit_criterion=[TimeBudget(1)], synchronous=True)
        """
        from ads.hpo.tuner_artifact import DownloadTunerArtifact

        tuner_args, cls.metadata = DownloadTunerArtifact(
            file_uri, target_file_path=target_file_path
        ).extract_tuner_args(delete_zip_file=delete_zip_file)
        return cls(**tuner_args)

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def _plot(
        self,  # type: ADSTuner
        plot_module,  # type: str
        plot_func,  # type: str
        time_interval=0.5,  # type: float
        fig_size=(800, 500),  # type: tuple
        **kwargs,
    ):
        if fig_size:
            logger.warning(
                "The param fig_size will be depreciated in future releases.",
            )

        spec = importlib.util.spec_from_file_location(
            "plot",
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "visualization",
                plot_module + ".py",
            ),
        )
        plot = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(plot)

        _imports.check()
        assert self._study is not None, "Need to call <code>.tune()</code> first."
        ntrials = 0
        if plot_func == "_plot_param_importances":
            print("Waiting for more trials before evaluating the param importance.")
        while self.status == State.RUNNING:
            import time
            from IPython.display import clear_output

            time.sleep(time_interval)
            if len(self.trials[~self.trials["value"].isnull()]) > ntrials:
                if plot_func == "_plot_param_importances":
                    if len(self.trials[~self.trials["value"].isnull()]) >= 4:
                        clear_output(wait=True)
                        getattr(plot, plot_func)(
                            study=self._study, fig_size=fig_size, **kwargs
                        )
                        clear_output(wait=True)
                else:
                    getattr(plot, plot_func)(
                        study=self._study, fig_size=fig_size, **kwargs
                    )
                    clear_output(wait=True)
            if len(self.trials) == 0:
                plt.figure()
                plt.title("Intermediate Values Plot")
                plt.xlabel("Step")
                plt.ylabel("Intermediate Value")
                plt.show(block=False)

            ntrials = len(self.trials[~self.trials["value"].isnull()])
        getattr(plot, plot_func)(study=self._study, fig_size=fig_size, **kwargs)

    def plot_best_scores(
        self,
        best=True,  # type: bool
        inferior=True,  # type: bool
        time_interval=1,  # type: float
        fig_size=(800, 500),  # type: tuple
    ):
        """Plot optimization history of all trials in a study.

        Parameters
        ----------
        best:
            controls whether to plot the lines for the best scores so far.
        inferior:
            controls whether to plot the dots for the actual objective scores.
        time_interval:
            how often(in seconds) the plot refresh to check on the new trial results.
        fig_size: tuple
            width and height of the figure.

        Returns
        -------
        None
            Nothing.
        """
        self._plot(
            "_optimization_history",
            "_get_optimization_history_plot",
            time_interval=time_interval,
            fig_size=fig_size,
            best=best,
            inferior=inferior,
        )

    @runtime_dependency(module="optuna", install_from=OptionalDependency.OPTUNA)
    def plot_param_importance(
        self,
        importance_evaluator="Fanova",  # type: str
        time_interval=1,  # type: float
        fig_size=(800, 500),  # type: tuple
    ):
        """Plot hyperparameter importances.

        Parameters
        ----------
        importance_evaluator: str
            Importance evaluator. Valid values: "Fanova", "MeanDecreaseImpurity". Defaults
            to "Fanova".
        time_interval: float
            How often the plot refresh to check on the new trial results.
        fig_size: tuple
            Width and height of the figure.

        Raises
        ------
        :class:`NotImplementedErorr`
            Raised for unsupported importance evaluators

        Returns
        -------
        None
            Nothing.
        """
        assert importance_evaluator in [
            "MeanDecreaseImpurity",
            "Fanova",
        ], "Only support <code>MeanDecreaseImpurity</code> and <code>Fanova</code>."
        if importance_evaluator == "Fanova":
            evaluator = None
        elif importance_evaluator == "MeanDecreaseImpurity":
            evaluator = optuna.importance.MeanDecreaseImpurityImportanceEvaluator()
        else:
            raise NotImplemented(
                f"{importance_evaluator} is not supported. It can be either `Fanova` or `MeanDecreaseImpurity`."
            )
        try:
            self._plot(
                plot_module="_param_importances",
                plot_func="_plot_param_importances",
                time_interval=time_interval,
                fig_size=fig_size,
                evaluator=evaluator,
            )
        except:
            logger.error(
                msg="""Cannot calculate the hyperparameter importance. Increase the number of trials or time budget. """
            )

    def plot_intermediate_scores(
        self,
        time_interval=1,  # type: float
        fig_size=(800, 500),  # type: tuple
    ):
        """
        Plot intermediate values of all trials in a study.

        Parameters
        ----------
        time_interval: float
            Time interval for the plot. Defaults to 1.
        fig_size: tuple[int, int]
            Figure size. Defaults to (800, 500).

        Returns
        -------
        None
            Nothing.
        """
        if not self._enable_pruning:
            logger.error(
                msg="Pruning was not used during tuning. "
                "There are no intermediate values to plot."
            )

        self._plot(
            "_intermediate_values",
            "_get_intermediate_plot",
            time_interval=time_interval,
            fig_size=fig_size,
        )

    def plot_edf_scores(
        self,
        time_interval=1,  # type: float
        fig_size=(800, 500),  # type: tuple
    ):
        """
        Plot the EDF (empirical distribution function) of the scores.

        Only completed trials are used.

        Parameters
        ----------
        time_interval: float
            Time interval for the plot. Defaults to 1.
        fig_size: tuple[int, int]
            Figure size. Defaults to (800, 500).

        Returns
        -------
        None
            Nothing.
        """
        self._plot(
            "_edf", "_get_edf_plot", time_interval=time_interval, fig_size=fig_size
        )

    def plot_contour_scores(
        self,
        params=None,  # type: Optional[List[str]]
        time_interval=1,  # type: float
        fig_size=(800, 500),  # type: tuple
    ):
        """
        Contour plot of the scores.

        Parameters
        ----------
        params: Optional[List[str]]
            Parameter list to visualize. Defaults to all.
        time_interval: float
            Time interval for the plot. Defaults to 1.
        fig_size: tuple[int, int]
            Figure size. Defaults to (800, 500).

        Returns
        -------
        None
            Nothing.
        """
        validate_params_for_plot(params, self._param_distributions)
        try:
            self._plot(
                "_contour",
                "_get_contour_plot",
                time_interval=time_interval,
                fig_size=fig_size,
                params=params,
            )
        except ValueError:
            logger.warning(
                msg="Cannot plot contour score."
                " Increase the number of trials or time budget."
            )

    def plot_parallel_coordinate_scores(
        self,
        params=None,  # type: Optional[List[str]]
        time_interval=1,  # type: float
        fig_size=(800, 500),  # type: tuple
    ):
        """
        Plot the high-dimentional parameter relationships in a study.

        Note that, If a parameter contains missing values, a trial with missing values is not plotted.

        Parameters
        ----------
        params: Optional[List[str]]
            Parameter list to visualize. Defaults to all.
        time_interval: float
            Time interval for the plot. Defaults to 1.
        fig_size: tuple[int, int]
            Figure size. Defaults to (800, 500).

        Returns
        -------
        None
            Nothing.
        """
        validate_params_for_plot(params, self._param_distributions)
        self._plot(
            "_parallel_coordinate",
            "_get_parallel_coordinate_plot",
            time_interval=time_interval,
            fig_size=fig_size,
            params=params,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
import functools
import nltk
import pandas as pd
from nltk import FreqDist, ngrams, pos_tag
from nltk.stem import SnowballStemmer
from nltk.tokenize import sent_tokenize, word_tokenize
from ads.feature_engineering.feature_type.adsstring.parsers.base import Parser
from typing import List

if "CONDA_PREFIX" in os.environ:
    nltk.data.path.append(f"{os.environ['CONDA_PREFIX']}/nltk")


class NLTKParser(Parser):

    stemmers = {}  # language -> stemmer

    def __init__(self):
        if not self.language in NLTKParser.stemmers:
            NLTKParser.stemmers[self.language] = SnowballStemmer(self.language)

    @property
    @functools.lru_cache()
    def pos(self) -> pd.DataFrame:
        return pd.DataFrame(data=self._pos(), columns=["Word", "Label"])

    @functools.lru_cache()
    def _pos(self):
        return pos_tag(word_tokenize(self.string))

    @property
    @functools.lru_cache()
    def adjective(self) -> List[str]:
        return [word for (word, pos) in self._pos() if pos.startswith("JJ")]

    @property
    @functools.lru_cache()
    def noun(self) -> List[str]:
        return [word for (word, pos) in self._pos() if pos.startswith("NN")]

    @property
    @functools.lru_cache()
    def adverb(self) -> List[str]:
        return [word for (word, pos) in self._pos() if pos.startswith("RB")]

    @property
    @functools.lru_cache()
    def verb(self) -> List[str]:
        return [word for (word, pos) in self._pos() if pos.startswith("VB")]

    @property
    @functools.lru_cache()
    def sentence(self) -> List[str]:
        return sent_tokenize(self.string)

    @property
    @functools.lru_cache()
    def token(self) -> List[str]:
        return word_tokenize(self.string)

    @property
    def word(self) -> List[str]:
        return [word.lower() for word in self.token if word.isalpha()]

    @property
    @functools.lru_cache()
    def word_count(self) -> pd.DataFrame:
        word_count = list(FreqDist(self.word).items())
        df = pd.DataFrame(data=word_count, columns=["Word", "Count"])
        return df.sort_values(
            ["Count", "Word"], ascending=[False, True], ignore_index=True
        )

    @property
    @functools.lru_cache()
    def bigram(self) -> pd.DataFrame:
        res = list(ngrams(self.word, 2))
        return pd.DataFrame(data=res, columns=["Word 1", "Word 2"])

    @property
    @functools.lru_cache()
    def trigram(self) -> pd.DataFrame:
        res = list(ngrams(self.word, 3))
        return pd.DataFrame(data=res, columns=["Word 1", "Word 2", "Word 3"])

    @property
    @functools.lru_cache()
    def stem(self) -> List:
        return [NLTKParser.stemmers[self.language].stem(w) for w in self.word]



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that helps to minimize the number of errors of the model post-deployment process.
The model provides a simple testing harness to ensure that model artifacts are
thoroughly tested before being saved to the model catalog.

Classes
--------
ModelIntrospect
    Class to introspect model artifacts.

Examples
--------
>>> model_introspect = ModelIntrospect(artifact=model_artifact)
>>> model_introspect()
... Test key         Test name            Result              Message
... ----------------------------------------------------------------------------
... test_key_1       test_name_1          Passed              test passed
... test_key_2       test_name_2          Not passed          some error occured
>>> model_introspect.status
... Passed
"""
from enum import Enum
import errno
import importlib
import json
import os
from abc import ABC
from copy import copy
from dataclasses import dataclass
from typing import List

import pandas as pd
from ads.common.model_metadata import MetadataTaxonomyKeys


class IntrospectionNotPassed(ValueError):
    pass


class TEST_STATUS(str):
    PASSED = "Passed"
    NOT_PASSED = "Failed"
    NOT_TESTED = "Skipped"


_PATH_TO_MODEL_ARTIFACT_VALIDATOR = (
    "ads.model_artifact_boilerplate.artifact_introspection_test.model_artifact_validate"
)
_INTROSPECT_METHOD_NAME = "validate_artifact"
_INTROSPECT_RESULT_FILE_NAME = "test_json_output.json"


class _PRINT_COLUMNS(Enum):
    KEY = "Test key"
    CASE = "Test name"
    RESULT = "Result"
    MESSAGE = "Message"


class _TEST_COLUMNS(str):
    CATEGORY = "category"
    DESCRIPTION = "description"
    ERROR_MSG = "error_msg"
    SUCCESS = "success"


_TEST_STATUS_MAP = {
    True: TEST_STATUS.PASSED,
    False: TEST_STATUS.NOT_PASSED,
    None: TEST_STATUS.NOT_TESTED,
}


class _ERROR_MESSAGES(str):
    MODEL_ARTIFACT_NOT_SET = "A model artifact is required."
    MODEL_ARTIFACT_INVALID_TYPE = (
        "The model artifact must be an instance of the class ModelArtifact."
    )


@dataclass
class PrintItem:
    """Class represents the model introspection print item."""

    key: str = ""
    case: str = ""
    result: str = ""
    message: str = ""

    def to_list(self) -> List[str]:
        """Converts instance to a list representation.

        Returns
        -------
        List[str]
            The instance in a list representation.
        """
        return [self.key, self.case, self.result, self.message]


class Introspectable(ABC):
    """Base class that represents an introspectable object."""

    pass


class ModelIntrospect:
    """Class to introspect model artifacts.

    Parameters
    ----------
    status: str
        Returns the current status of model introspection.
        The possible variants: `Passed`, `Not passed`, `Not tested`.
    failures: int
        Returns the number of failures of introspection result.

    Methods
    -------
    run(self) -> None
        Invokes model artifacts introspection.
    to_dataframe(self) -> pd.DataFrame
        Serializes model introspection result into a DataFrame.

    Examples
    --------
    >>> model_introspect = ModelIntrospect(artifact=model_artifact)
    >>> result = model_introspect()
    ... Test key         Test name            Result              Message
    ... ----------------------------------------------------------------------------
    ... test_key_1       test_name_1          Passed              test passed
    ... test_key_2       test_name_2          Not passed          some error occured
    """

    def __init__(self, artifact: Introspectable):
        """Initializes the Model Introspect.

        Parameters
        ----------
        artifact: Introspectable
            The instance of ModelArtifact object.

        Raises
        ------
            ValueError: If model artifact object not provided.
            TypeError: If provided input paramater not a ModelArtifact instance.
        """
        if not artifact:
            raise ValueError(_ERROR_MESSAGES.MODEL_ARTIFACT_NOT_SET)

        if not isinstance(artifact, Introspectable):
            raise TypeError(_ERROR_MESSAGES.MODEL_ARTIFACT_INVALID_TYPE)

        self._artifact = artifact
        self._reset()

    def _reset(self) -> None:
        """Resets test result to initial state."""
        self._status = TEST_STATUS.NOT_TESTED
        self._result = None
        self._prepared_result = []

    def _save_result_to_artifacts(self) -> None:
        """Saves introspection result into the model artifacts folder.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
            FileNotFoundError: If path to model artifacts does not exist.
        """
        if not os.path.isdir(self._artifact.artifact_dir):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), self._artifact.artifact_dir
            )

        output_file = f"{self._artifact.artifact_dir}/{_INTROSPECT_RESULT_FILE_NAME}"
        with open(output_file, "w") as f:
            json.dump(self._result, f, indent=4)

    def _save_result_to_metadata(self) -> None:
        """Saves the result of introspection to the model metadata."""
        self._artifact.metadata_taxonomy[
            MetadataTaxonomyKeys.ARTIFACT_TEST_RESULT
        ].update(value=self._result)

    def _import_and_run_validator(self) -> None:
        """Imports and run model artifact validator.

        The validator provided as one of the modules of model artifacts boilerplate.
        The importlib API is used to load validator and to invoke test method.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        FileNotFoundError: If path to model artifacts does not exist.
        """

        if not os.path.isdir(self._artifact.artifact_dir):
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
            )

        module = importlib.import_module(_PATH_TO_MODEL_ARTIFACT_VALIDATOR)
        importlib.reload(module)
        method = getattr(module, _INTROSPECT_METHOD_NAME)
        params = {"artifact": self._artifact.artifact_dir}
        test_result, _ = method(**params)

        self._status = _TEST_STATUS_MAP.get(test_result)
        self._result = copy(module.TESTS)
        self._prepared_result = self._prepare_result()

    @property
    def status(self) -> str:
        """Gets the current status of model introspection."""
        return self._status

    def run(self) -> pd.DataFrame:
        """Invokes introspection.

        Returns
        -------
        pd.DataFrame
           The introspection result in a DataFrame format.
        """
        self._reset()
        self._import_and_run_validator()
        self._save_result_to_metadata()
        self._save_result_to_artifacts()
        return self.to_dataframe()

    def _prepare_result(self) -> List[PrintItem]:
        """Prepares introspection result information to display to user.

        Returns
        -------
        List[PrintItem]
            The list of prepared to print data items.
        """
        if not self._result:
            return []

        result = []
        for key, item in self._result.items():
            error_msg = (
                item.get(_TEST_COLUMNS.ERROR_MSG)
                if (
                    item.get(_TEST_COLUMNS.SUCCESS) == False
                    or (
                        item.get(_TEST_COLUMNS.SUCCESS) == None
                        and "WARNING" in item.get(_TEST_COLUMNS.ERROR_MSG, "")
                    )
                )
                else ""
            )
            result.append(
                PrintItem(
                    key,
                    item.get(_TEST_COLUMNS.DESCRIPTION, ""),
                    _TEST_STATUS_MAP.get(item.get(_TEST_COLUMNS.SUCCESS)),
                    error_msg,
                )
            )
        return result

    def to_dataframe(self) -> pd.DataFrame:
        """Serializes model introspection result into a DataFrame.

        Returns
        -------
        `pandas.DataFrame`
            The model introspection result in a DataFrame representation.
        """
        return (
            pd.DataFrame(
                (item.to_list() for item in self._prepared_result),
                columns=[item.value for item in _PRINT_COLUMNS],
            )
            .sort_values(by=[_PRINT_COLUMNS.KEY.value, _PRINT_COLUMNS.CASE.value])
            .reset_index(drop=True)
        )

    @property
    def failures(self) -> int:
        """Calculates the number of failures.

        Returns
        -------
        int
            The number of failures.
        """
        return len(
            [
                item
                for item in self._prepared_result
                if item.result == TEST_STATUS.NOT_PASSED
            ]
        )

    def __call__(self) -> pd.DataFrame:
        """Invokes introspection.

        Returns
        -------
        pd.DataFrame
           The introspection result in a DataFrame format.
        """
        return self.run()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from time import sleep, time_ns
from ads.opctl.distributed.common.cluster_provider_factory import ClusterProviderFactory
import traceback


class ClusterRunner:
    def __init__(self, cluster_provider=None, cluster_key=None):
        self.cluster_key = cluster_key or os.environ.get("OCI__CLUSTER_TYPE")
        self.mode = os.environ.get("OCI__MODE")
        self.ephemeral = os.environ.get("OCI__EPHEMERAL", 1)
        # life_span = os.environ.get("OCI__LIFE_SPAN")  # TODO: Ask MR how this works
        self.work_dir = os.environ.get("OCI__WORK_DIR")
        os.environ["JOB_OCID"] = os.environ.get("JOB_OCID", 'Undefined')
        os.environ["JOB_RUN_OCID"] = os.environ.get("JOB_RUN_OCID", str(time_ns()))
        self.cluster = cluster_provider or ClusterProviderFactory.get_provider(
            self.cluster_key,
            mode=self.mode,
            ephemeral=self.ephemeral,
            work_dir=self.work_dir,
        )  # life_spanlife_span=life_span
        print(f"Cluster built: {self.cluster}", flush=True)

    def run(self):
        exit_code = 0
        self.cluster.start()
        try:
            self.cluster.run_code()
            # self.cluster.code_execution_complete = True # This needs to be
            # set inside the run_code method of the implementation class.
        except Exception as e:
            print(f"Error Running the code: {e}", flush=True)
            traceback.print_exc()
            exit_code = 1
            self.cluster.execution_failed()
        while (
            not self.cluster.tearable()
        ):  # If not ephemeral, wait util it is ready for tearing down
            sleep(15)
        print("Signalling Stop!!!", flush=True)
        self.cluster.stop()  # Signal cluster tear down
        print(f"Exiting with exit code: {exit_code}", flush=True)
        self.cluster.sync(loop=False)
        exit(exit_code)


if __name__ == "__main__":
    ClusterRunner().run()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import yaml
import os
from jinja2 import Environment, PackageLoader

# FunctionAttributes = namedtuple('FunctionAttributes',['schema_version','name','version','runtime','entrypoint','memory'])
fn_config = None
env = Environment(loader=PackageLoader("ads", "templates"))


def prepare_fn_attributes(
    func_name: str,
    schema_version=20180708,
    version=None,
    python_runtime=None,
    entry_point=None,
    memory=None,
) -> dict:
    """
    Workaround for collections.namedtuples. The defaults are not supported.
    """
    function_attributes = {}
    function_attributes["schema_version"] = schema_version
    function_attributes["name"] = func_name
    function_attributes["version"] = (
        version if version else get_function_config()["fn_conf"]["version"]
    )
    function_attributes["runtime"] = (
        python_runtime
        if python_runtime
        else get_function_config()["fn_conf"]["runtime"]
    )
    function_attributes["entrypoint"] = (
        entry_point if entry_point else get_function_config()["fn_conf"]["entrypoint"]
    )
    function_attributes["memory"] = (
        memory if memory else get_function_config()["fn_conf"]["memory"]
    )

    return function_attributes


def get_function_config() -> dict:
    """
    Returns dictionary loaded from func_conf.yaml
    """
    global fn_config
    if fn_config:
        return fn_config
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "func_conf.yaml")
    ) as fn_config_file:
        func_config = yaml.load(fn_config_file, Loader=yaml.FullLoader)
    return func_config


def generate_fn_artifacts(
    path: str,
    fn_name: str = None,
    fn_attributes=None,
    artifact_type_generic=False,
    **kwargs,
):
    """
    Generates artifacts for fn (https://fnproject.io) at the provided path -
        * func.py
        * func.yaml
        * requirements.txt if not there. If exists appends fdk to the file.
        * score.py

    Parameters
    ----------
        path: str
            Target folder where the artifacts are placed.
        fn_attributes: dict
            dictionary specifying all the function attributes as described in https://github.com/fnproject/docs/blob/master/fn/develop/func-file.md
        artifact_type_generic: bool
            default is False. This attribute decides which template to pick for score.py. If True, it is assumed that the code to load is provided by the user.

    """
    assert fn_name or fn_attributes, (
        "Must provide either fn_name or fn_attributes. You may use "
        "ads.common.artifact.fn_util.prepare_fn_attributes for creating fn_attributes "
    )
    function_attributes = prepare_fn_attributes(fn_name) if fn_name else fn_attributes

    progress = kwargs.get("progress", None)
    if progress:
        progress.update("Writing func.yaml")
    with open(os.path.join(path, "func.yaml"), "w") as fnyaml_file:
        yaml.dump(function_attributes, fnyaml_file)

    if progress:
        progress.update("Writing func.py")
    func_template = env.get_template("func.jinja2")
    with open(os.path.join(path, "func.py"), "w") as func_fl:
        func_fl.write(func_template.render(score_module="score"))


def write_score(path, **kwargs):
    serializer = kwargs.get("serializer", "default")
    if serializer not in get_function_config()["models"]:
        serializer = "default"

    model_name = kwargs.get(
        "model_name", get_function_config()["models"][serializer]["file_name"]
    )
    transformer_name = (
        get_function_config()["models"][serializer]["transformer_name"]
        if serializer in get_function_config()["models"]
        else "onnx_data_transformer.json"
    )
    misc_imports = (
        kwargs.get("misc_imports", list())
        + get_function_config()["models"][serializer]["misc_imports"]
    )
    underlying_model = kwargs.get("underlying_model", "NOTFOUND")
    sklearn_model = underlying_model in ["sklearn", "xgboost"]

    if underlying_model == "automl":
        jinja_template_filename = "score_oracle_automl"
    else:
        jinja_template_filename = kwargs.get(
            "input_file",
            get_function_config()["models"][serializer].get("input_file", "score"),
        )
    scorefn_template = env.get_template(f"{jinja_template_filename}.jinja2")
    with open(os.path.join(path, "score.py"), "w") as sfl:
        sfl.write(
            scorefn_template.render(
                model_file_name=model_name,
                transfromer_file_name=transformer_name,
                misc_imports=misc_imports,
                sklearn_model=sklearn_model,
                underlying_model_type=kwargs.get("_underlying_model", None),
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class DatasetError(BaseException):
    """Base class for dataset errors."""

    def __init__(self, *args, **kwargs):
        pass

    def __str__(self, *args, **kwargs):
        pass


class ValidationError(DatasetError):
    """Handles validation errors in dataset."""

    def __init__(self, msg):
        self.msg = msg

    def __str__(self, *args, **kwargs):
        return self.msg
